"""Records queue and Record work object for queue."""
import contextlib
import json
import os
from pathlib import Path
from queue import Queue
from typing import Any, Optional

from rich import print  # pylint: disable=redefined-builtin

from histdatacom.fx_enums import Format, Timeframe
from histdatacom.utils import (
    create_full_path,
    get_query_string,
)


class Record:  # noqa:H601
    """A work record for the queue."""

    def __init__(self, **kwargs: str) -> None:
        """Initialize record attributes.

        Args:
            kwargs (str): record attributes
        """
        self.url = kwargs.get("url", "")
        self.status = kwargs.get("status", "")
        self.encoding = kwargs.get("encoding", "")
        self.bytes_length = kwargs.get("bytes_length", "")
        self.data_date = kwargs.get("data_date", "")
        self.data_year = kwargs.get("data_year", "")
        self.data_month = kwargs.get("data_month", "")
        self.data_datemonth = kwargs.get("data_datemonth", "")
        self.data_format = kwargs.get("data_format", "")
        self.data_timeframe = kwargs.get("data_timeframe", "")
        self.data_fxpair = kwargs.get("data_fxpair", "")
        self.data_dir = kwargs.get("data_dir", "")
        self.data_tk = kwargs.get("data_tk", "")
        self.zip_filename = kwargs.get("zip_filename", "")
        self.csv_filename = kwargs.get("csv_filename", "")
        self.jay_filename = kwargs.get("jay_filename", "")
        self.jay_line_count = kwargs.get("jay_line_count", "")
        self.jay_start = kwargs.get("jay_start", "")
        self.jay_end = kwargs.get("jay_end", "")
        self.zip_persist = kwargs.get("zip_persist", "")

    def write_memento_file(self, base_dir: str = "") -> None:
        """Write record to disk.

        # noqa: DAR402

        Args:
            base_dir (str): Defaults to "".

        Raises:
            ValueError: function requires base_dir.
            SystemExit: Exit on error.
        """
        try:
            if self.data_dir == "":
                if base_dir:
                    self._create_record_data_dir(base_dir=base_dir)
                else:
                    raise ValueError

            if not Path(self.data_dir).exists():
                create_full_path(self.data_dir)

            momento_path = Path(self.data_dir, ".meta")

            with momento_path.open("w", encoding="UTF-8") as target:
                json.dump(self._to_dict(), target)

        except ValueError as err:
            print(  # noqa:T201,BLK100
                "Error: create_record_data_dir not provided base_dir="
            )
            raise SystemExit from err

    def delete_momento_file(self) -> None:
        """Delete memento file."""
        momento_path = Path(self.data_dir, ".meta")
        if momento_path.exists():
            momento_path.unlink()

    def restore_momento(self, base_dir: str) -> bool:
        """Restore momento from .meta file.

        Args:
            base_dir (str): base data directory

        Returns:
            bool: True (success) | False (failure)
        """
        self._set_record_data_dir(base_dir)

        momento_path = Path(self.data_dir, ".meta")
        if not momento_path.exists():
            return False
        record_dict: dict = {}

        with (
            momento_path.open(  # noqa:BLK100
                "r", encoding="UTF-8"
            ) as json_read,
            contextlib.suppress(Exception),
        ):
            while True:
                record_dict |= json.load(json_read)

        self(**record_dict)
        return True

    def _to_dict(self) -> dict:
        """Return dict representation of Record.

        Returns:
            dict: dict representation of Record.
        """
        return {
            "url": self.url,
            "status": self.status,
            "encoding": self.encoding,
            "bytes_length": self.bytes_length,
            "data_date": self.data_date,
            "data_year": self.data_year,
            "data_month": self.data_month,
            "data_datemonth": self.data_datemonth,
            "data_format": self.data_format,
            "data_timeframe": self.data_timeframe,
            "data_fxpair": self.data_fxpair,
            "data_dir": self.data_dir,
            "data_tk": self.data_tk,
            "zip_filename": self.zip_filename,
            "csv_filename": self.csv_filename,
            "jay_line_count": self.jay_line_count,
            "jay_start": self.jay_start,
            "jay_end": self.jay_end,
            "jay_filename": self.jay_filename,
            "zip_persist": self.zip_persist,
        }

    def _set_record_data_dir(  # noqa:CFQ004,BLK100
        self, base_dir: Optional[str]
    ) -> str:
        """Set Record's data directory.

        Args:
            base_dir (Optional[str]): base data directory.

        Returns:
            str: self.data_dir  # record's data.
        """
        query_string_args = get_query_string(self.url)
        length = len(query_string_args)

        csv_format = Format(query_string_args[1]).name
        timeframe = Timeframe(query_string_args[2]).name

        record_data_dir = f"{base_dir}{csv_format}{os.sep}{timeframe}{os.sep}"

        if length == 3:
            self.data_dir = record_data_dir
            return self.data_dir

        pair = query_string_args[3]
        record_data_dir = record_data_dir + pair.lower() + os.sep

        if length == 4:
            self.data_dir = record_data_dir
            return self.data_dir

        year = query_string_args[4]
        record_data_dir = record_data_dir + year + os.sep

        if length == 5:
            self.data_dir = record_data_dir
            return self.data_dir

        month = query_string_args[5]
        record_data_dir = record_data_dir + month + os.sep

        if length == 6:
            self.data_dir = record_data_dir
            return self.data_dir

        return self.data_dir

    def _create_record_data_dir(self, base_dir: str = "") -> None:
        """Create Record's data directory and populate its attribute.

        # noqa: DAR402

        Args:
            base_dir (str): Defaults to "".

        Raises:
            ValueError: no base_dir provided.
            SystemExit: exit on error.
        """
        try:
            if self.data_dir != "":
                create_full_path(self.data_dir)
            elif base_dir != "":
                create_full_path(self._set_record_data_dir(base_dir))
            else:
                raise ValueError
        except ValueError as err:
            print(  # noqa:BLK100,T201
                "Error: create_record_data_dir not provided base_dir="
            )
            raise SystemExit from err

    def __call__(self, **kwargs: str) -> Any:
        """Set instance attribute by kwargs.

        Args:
            kwargs (str): instance attributes

        Returns:
            Any: self.
        """
        for arg_name, arg_value in kwargs.items():
            setattr(self, arg_name, arg_value)
        return self


class Records(Queue):  # noqa:H601
    """Custom Queue class for Records and SyncManager."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Inherit and initialize from queue.Queue.

        Args:
            args (Any): for compatibility with queue.Queue
            kwargs (Any): for compatibility with queue.Queue
        """
        Queue.__init__(self, *args, **kwargs)

    def dump_to_queue(  # noqa:CCR001
        self, dst_queue: Queue, count: int | None = 0
    ) -> None:
        """Transfer queue contents from one queue to another.

        Args:
            dst_queue (Queue): destination queue
            count (int): Dump N number of records. Defaults to 0.
        """
        if count == 0:
            count = None
        else:
            counter = 0

        while not self.empty():
            record = self.get()

            if record is None:
                break

            if count is None:
                dst_queue.put(record)
            elif counter < count:
                dst_queue.put(record)
                counter += 1
            else:
                self.put(record)
                break



"""Command-Line interface for histdatacom.

CLI interface for histdatacom

Raises:
    ValueError: ERROR on -f {self.arg_namespace.formats}           ERROR
                * format must be ASCII when importing to influxdb. eg. -f ascii
    ValueError: ERROR on -e start           ERROR
                                    * keyword 'start' cannot be used as -e start
    ValueError: ERROR on -s start           ERROR
                                    * keyword 'start' must also specify
                                      an end year-month
    ValueError: ERROR on -e {end_yearmonth}  ERROR
                    * Malformed command:
                        - cannot include `-e end_year-month` without
                          specifying a start year-month.
                            eg. -s year-month -e year-month
    ValueError: ERROR on:
            -e {get_year_from_datemonth(start_yearmonth)}  {end_yearmonth}  ERROR
                    * Malformed command:
                        - cannot include `-e end_year-month` when
                          specifying a single year
                          with -s {get_year_from_datemonth(start_yearmonth)}
    ValueError: ERROR on -e {start_yearmonth}   ERROR
                    start month cannot be zero.
                        * valid inputs:
                            a) just the year
                                eg. -s 2022
                            b) months 1-12:
                                eg. -s 2022-04
    ValueError: ERROR on -s {start_yearmonth}  ERROR
                    * Malformed command:
                        - start month is greater than 12.
                          valid input is 01-12.
    ValueError: ERROR on -e {end_yearmonth}           ERROR
                            * You left out the end month.
                                - valid input is -e year-month(1-12)
                                    eg. -e 2022-03
    ValueError: ERROR on -e {end_yearmonth}           ERROR
                            * End month cannot be zero.
                                - valid input is -e year-month(1-12)
                                    eg. -e 2022-03
    ValueError: ERROR on -e {end_yearmonth}  ERROR
                            * Malformed command:
                                - end month is greater than 12.
                                valid input is 01-12.
    ValueError: ERROR on -s {start_yearmonth} -e {end_yearmonth}  ERROR
                    * start year-month and end year-month cannot be the same.
    ValueError: ERROR on -s {start_yearmonth}      ERROR
                            * bad year-month
                                - no data available for dates
                                prior to 2000y
    ValueError: ERROR on -s {start_yearmonth}      ERROR
                            * year-month cannot be in the future.
    ValueError: ERROR on -e {end_yearmonth}     ERROR
                            * bad year-month
                                - no data available for dates
                                prior to 2000y
    ValueError: ERROR on -e {end_yearmonth}     ERROR
                            * year-month cannot be in the future.
    ValueError: ERROR on -s {start_yearmonth} -e {end_yearmonth}    ERROR
                        * logic error: end year-month is before start year-month.
    ValueError: ERROR on {yearmonth}    ERROR
                            * invalid yearmonth format

Returns:
    _type_: self.arg_namespace
"""

import argparse
import re
import sys
from typing import Any, Optional, Tuple

from rich import print  # pylint: disable=redefined-builtin

from histdatacom import Options
from histdatacom.concurrency import get_pool_cpu_count
from histdatacom.fx_enums import Format, Pairs, Timeframe
from histdatacom.utils import (
    get_current_datemonth_gmt_minus5,
    get_year_from_datemonth,
    get_month_from_datemonth,
    replace_date_punct,
)

default_options = Options()


class ArgParser(argparse.ArgumentParser):  # noqa:H601
    """Encapsulation class for argparse related operations."""

    def __init__(self, options: Options = default_options):
        """Set up argparse.

        bring in defaults DTO, setup cli params, receive
        and overwrite defaults with user cli args.

        Args:
            options (Options): _description_. Defaults to default_options.
        """
        # init _HistDataCom.ArgParser to extend argparse.ArgumentParser
        argparse.ArgumentParser.__init__(self, prog="histdatacom")
        # bring in the defaults arg DTO from outer class, use the
        # __dict__ representation of it to set argparse argument defaults.

        self.arg_namespace = options
        self._default_args = self.arg_namespace.__dict__
        self.set_defaults(**self._default_args)

    @classmethod
    def arg_list_to_set(cls, args: dict) -> dict:
        """Convert any lists in from argparse to sets.

         This is to standardize data types. If the user specifies a parameter,
         argparse returns a list, our defaults are sets, so .

        Args:
            args (dict): argparse dict

        Returns:
            dict: args dict
        """
        for arg in args:
            if isinstance(args[arg], list):
                args[arg] = set(args[arg])
        return args

    def _adjust_for_repo_data_request(self) -> None:
        """Null user inputs when -A or -U are set and ready for repo run."""
        if (
            self.arg_namespace.available_remote_data
            or self.arg_namespace.update_remote_data
        ):
            self.arg_namespace.start_yearmonth = None
            self.arg_namespace.end_yearmonth = None
            self.arg_namespace.validate_urls = False
            self.arg_namespace.download_data_archives = False
            self.arg_namespace.extract_csvs = False
            self.arg_namespace.import_to_influxdb = False
            self.arg_namespace.formats = {"ascii"}
            self.arg_namespace.timeframes = {"tick-data-quotes"}

    def _clean_from_api_args(self) -> list:  # noqa:CCR001
        """Build the args list from api Options.

        Returns:
            list: args
        """
        self.arg_namespace.timeframes = Timeframe.convert_to_values(
            self.arg_namespace.timeframes
        )

        args = [
            *["--data-directory", self.arg_namespace.data_directory],
            *["-p", *self.arg_namespace.pairs],
            *["-f", *self.arg_namespace.formats],
            *["-t", *self.arg_namespace.timeframes],
            *["-c", self.arg_namespace.cpu_utilization],
            *["-b", self.arg_namespace.batch_size],
        ]

        if self.arg_namespace.start_yearmonth:
            args.extend(["-s", self.arg_namespace.start_yearmonth])
        if self.arg_namespace.end_yearmonth:
            args.extend(["-e", self.arg_namespace.end_yearmonth])
        if self.arg_namespace.available_remote_data:
            args.append("-A")
        if self.arg_namespace.update_remote_data:
            args.append("-U")
        if self.arg_namespace.validate_urls:
            args.append("-V")
        if self.arg_namespace.download_data_archives:
            args.append("-D")
        if self.arg_namespace.extract_csvs:
            args.append("-X")
        if self.arg_namespace.import_to_influxdb:
            args.append("-I")
        if self.arg_namespace.delete_after_influx:
            args.append("-d")
        return args

    def _false_from_api_if_behavior_flag(self) -> None:
        """Null the API flag if a behavior flag has been set."""
        if (
            self.arg_namespace.validate_urls
            or self.arg_namespace.download_data_archives
            or self.arg_namespace.extract_csvs
            or self.arg_namespace.import_to_influxdb
        ):
            self.arg_namespace.from_api = False

    def _check_for_ascii_if_influx(self) -> None:
        """Verify ascii csv_format type for influxdb import.

        # noqa: DAR402

        Raises:
            ValueError: if influx, must use ascii
            SystemExit: exit on error
        """
        try:
            if self.arg_namespace.import_to_influxdb:
                err_text_influx_must_be_ascii = f"""
            ERROR on -f {self.arg_namespace.formats}           ERROR
                * format must be ASCII when importing to influxdb. eg. -f ascii

            """
                for csv_format in self.arg_namespace.formats:
                    if str.lower(csv_format) != "ascii":
                        raise ValueError(err_text_influx_must_be_ascii)
        except ValueError as err:
            print(err)  # noqa:T201
            raise SystemExit from err

    def _check_for_ascii_if_api(self) -> None:  # noqa:CCR001
        """Verify ascii csv_format type for api use.

        # noqa: DAR402

        Raises:
            ValueError: if influx, must use ascii
            SystemExit: exit on error
        """
        try:
            if self.arg_namespace.from_api and not (
                self.arg_namespace.validate_urls
                or self.arg_namespace.download_data_archives
                or self.arg_namespace.extract_csvs
                or self.arg_namespace.import_to_influxdb
            ):
                err_text_api_must_be_ascii = f"""
                ERROR on -f {self.arg_namespace.formats}           ERROR
                    * format must be ASCII when calling from API
                        eg.
                            import histdatacom
                            from histdatacom.options import Options

                            options = Options()
                            options.formats = {{"ascii"}}
            """
                for csv_format in self.arg_namespace.formats:
                    if str.lower(csv_format) != "ascii":
                        raise ValueError(err_text_api_must_be_ascii)
        except ValueError as err:
            print(err)  # noqa:T201
            raise SystemExit from err

    def _check_datetime_input(self) -> None:
        """Check for invalid datetime input for -s and -e flags."""
        if (  # noqa:BLK001
            self.arg_namespace.start_yearmonth  # noqa:BLK001
            or self.arg_namespace.end_yearmonth
        ):
            (
                self.arg_namespace.start_yearmonth,
                self.arg_namespace.end_yearmonth,
            ) = self._check_for_start_in_yearmonth()

            (
                self.arg_namespace.start_yearmonth,
                self.arg_namespace.end_yearmonth,
            ) = self._check_for_now_in_yearmonth()

            (
                self.arg_namespace.start_yearmonth,
                self.arg_namespace.end_yearmonth,
            ) = self._check_cli_start_yearmonth()

            self._check_cli_end_yearmonth()

            self._check_for_same_start_yearmonth()

        (
            self.arg_namespace.start_yearmonth,
            self.arg_namespace.end_yearmonth,
        ) = self._replace_falsey_yearmonth_with_none()

        self._check_start_yearmonth_in_range()
        self._check_end_yearmonth_in_range()
        self._check_start_lessthan_end()
        self._validate_prerequisites()

    def _validate_prerequisites(self) -> None:
        """Set prereqs for behavior flags -V -D -X -I."""
        if self.arg_namespace.validate_urls:
            return

        if self.arg_namespace.download_data_archives:
            self.arg_namespace.validate_urls = True

        if self.arg_namespace.extract_csvs:
            self.arg_namespace.validate_urls = True
            self.arg_namespace.download_data_archives = True

        if self.arg_namespace.import_to_influxdb:
            self.arg_namespace.validate_urls = True
            self.arg_namespace.download_data_archives = True

        if (
            not self.arg_namespace.download_data_archives
            and not self.arg_namespace.extract_csvs
            and not self.arg_namespace.import_to_influxdb
        ):
            self.arg_namespace.validate_urls = True
            self.arg_namespace.download_data_archives = True
            self.arg_namespace.extract_csvs = True

    def _check_for_now_in_yearmonth(  # noqa:CCR001
        self,
    ) -> Tuple[Optional[str], Optional[str]]:
        """Check for now in -s or -e and adjusts it to current year-month.

        Returns:
            Tuple[Optional[str], Optional[str]]: start_yearmonth, end_yearmonth
        """
        if start_yearmonth := self.arg_namespace.start_yearmonth:
            if start_yearmonth == "now":
                return get_current_datemonth_gmt_minus5(), None
            if end_yearmonth := self.arg_namespace.end_yearmonth:  # noqa:SIM102
                if end_yearmonth == "now":
                    return (
                        start_yearmonth,
                        get_current_datemonth_gmt_minus5(),
                    )

        return start_yearmonth, end_yearmonth

    def _check_for_start_in_yearmonth(  # noqa:CCR001
        self,
    ) -> Tuple[Optional[str], Optional[str]] | Any:
        """Check for 'start' keyword in -s and sets -s yearmonth to 200001.

        # noqa: DAR402

        Raises:
            ValueError: end_yearmonth cannot be "start"
            ValueError: keyword 'start' must also specify an end year-month.
            SystemExit: exit on error

        Returns:
            Tuple[Optional[str], Optional[str]] | Any: start_yearmonth,
                                                       end_yearmonth
        """
        try:
            if (start_yearmonth := self.arg_namespace.start_yearmonth) and (
                start_yearmonth == "start"
            ):
                if end_yearmonth := self.arg_namespace.end_yearmonth:
                    if end_yearmonth == "start":
                        err_text_end_yearmonth_cannot_be_start = """
                            ERROR on -e start           ERROR
                                * keyword 'start' cannot be used as -e start
                        """
                        raise ValueError(err_text_end_yearmonth_cannot_be_start)
                    return "200001", end_yearmonth

                err_text_start_must_have_end = """
                        ERROR on -s start           ERROR
                            * keyword 'start' must also specify
                                an end year-month
                """
                raise ValueError(err_text_start_must_have_end)  # noqa:TC301
            return (  # noqa:TC300
                self.arg_namespace.start_yearmonth,
                self.arg_namespace.end_yearmonth,
            )
        except ValueError as err:
            raise SystemExit from err

    def _check_cli_start_yearmonth(  # noqa:CCR001
        self,
    ) -> Tuple[Optional[str], Optional[str]]:
        """Validate for -s start_yearmonth.

        # noqa: DAR402

        Raises:
            ValueError: No start_yearmonth
            ValueError: No end_yearmonth
            ValueError: Start month cannot be zero.
            ValueError: Start month cannot be greater than 12.
            SystemExit: exit on error.

        Returns:
            Tuple[Optional[str], Optional[str]]: _description_
        """
        start_yearmonth = self.arg_namespace.start_yearmonth
        start_year = get_year_from_datemonth(start_yearmonth)
        start_month = get_month_from_datemonth(str(start_yearmonth))

        err_text_start_month = f"""
                ERROR on -e {start_yearmonth}   ERROR
                    start month cannot be zero.
                        * valid inputs:
                            a) just the year
                                eg. -s 2022
                            b) months 1-12:
                                eg. -s 2022-04
        """

        end_yearmonth = self.arg_namespace.end_yearmonth

        err_text_no_end_yearmonth = f"""
        ERROR on -e {get_year_from_datemonth(start_yearmonth)} {end_yearmonth}
            * Malformed command:
                - cannot include `-e end_year-month` when
                    specifying a single year
                    with -s {get_year_from_datemonth(start_yearmonth)}
        """

        err_text_no_start_yearmonth = f"""
                ERROR on -e {end_yearmonth}  ERROR
                    * Malformed command:
                        - cannot include `-e end_year-month` without
                          specifying a start year-month.
                            eg. -s year-month -e year-month
        """
        err_text_start_month_greater_than_12 = f"""
                ERROR on -s {start_yearmonth}  ERROR
                    * Malformed command:
                        - start month is greater than 12.
                          valid input is 01-12.
        """

        try:
            if not start_month:
                if end_yearmonth:
                    if not start_year:
                        raise ValueError(err_text_no_start_yearmonth)
                    raise ValueError(err_text_no_end_yearmonth)
                return f"{start_year}00", None
            if start_month == "00":
                raise ValueError(err_text_start_month)  # noqa:TC301
            if int(start_month) > 12:
                raise ValueError(  # noqa:TC301,BLK100
                    err_text_start_month_greater_than_12
                )
            return start_yearmonth, end_yearmonth  # noqa:TC300
        except ValueError as err:
            raise SystemExit from err

    def _check_cli_end_yearmonth(self) -> None:  # noqa:CCR001
        """Validate for -e end_yearmonth.

        # noqa: DAR402

        Raises:
            ValueError: no end month specified.
            ValueError: end month cannot be zero
            ValueError: end month cannot be greater than zero
            SystemExit: exit on error
        """
        try:
            if end_yearmonth := self.arg_namespace.end_yearmonth:
                end_year = get_year_from_datemonth(end_yearmonth)
                end_month = get_month_from_datemonth(end_yearmonth)

                err_text_no_endmonth = f"""
                        ERROR on -e {end_yearmonth}           ERROR
                            * You left out the end month.
                                - valid input is -e year-month(1-12)
                                    eg. -e 2022-03

                """

                err_text_endmonth_cannot_be_zero = f"""
                        ERROR on -e {end_yearmonth}           ERROR
                            * End month cannot be zero.
                                - valid input is -e year-month(1-12)
                                    eg. -e 2022-03

                """
                err_text_end_month_greater_than_12 = f"""
                        ERROR on -e {end_yearmonth}  ERROR
                            * Malformed command:
                                - end month is greater than 12.
                                valid input is 01-12.
                """
                if end_year and not end_month:
                    raise ValueError(err_text_no_endmonth)
                if end_month == "00":
                    raise ValueError(err_text_endmonth_cannot_be_zero)
                if int(end_month) > 12:
                    raise ValueError(err_text_end_month_greater_than_12)
        except ValueError as err:
            raise SystemExit from err

    def _check_for_same_start_yearmonth(self) -> None:
        """Validate -s start_yearmonth and -e end_yearmonth are not the same.

        # noqa: DAR402

        Raises:
            ValueError: -s and -e cannot be the same.
            SystemExit: Exit on error
        """
        try:
            start_yearmonth = self.arg_namespace.start_yearmonth
            start_year = get_year_from_datemonth(start_yearmonth)
            start_month = get_month_from_datemonth(str(start_yearmonth))

            end_yearmonth = self.arg_namespace.end_yearmonth
            end_year = get_year_from_datemonth(end_yearmonth)
            end_month = get_month_from_datemonth(str(end_yearmonth))

            err_text_start_and_end_cannot_be_the_same = f"""
                ERROR on -s {start_yearmonth} -e {end_yearmonth}  ERROR
                    * start year-month and end year-month cannot be the same.
            """

            if f"{start_year}_{start_month}" == f"{end_year}_{end_month}":
                raise ValueError(  # noqa:TC301
                    err_text_start_and_end_cannot_be_the_same
                )
        except ValueError as err:
            raise SystemExit from err

    def _replace_falsey_yearmonth_with_none(
        self,
    ) -> Tuple[Optional[str], Optional[str]]:
        """Set value to None if no user input on -s or -e.

        Returns:
            Tuple[Optional[str], Optional[str]]: start_yearmonth, end_yearmonth
        """
        start_yearmonth = self.arg_namespace.start_yearmonth
        end_yearmonth = self.arg_namespace.end_yearmonth

        if not start_yearmonth or start_yearmonth == "":
            start_yearmonth = None
        if not end_yearmonth or end_yearmonth == "":
            end_yearmonth = None

        return start_yearmonth, end_yearmonth

    def _check_start_yearmonth_in_range(self) -> None:
        """Validate -s not before than 200001 or later than current yearmonth.

        Raises:
            SystemExit: Exit on error
        """
        try:
            if start_yearmonth := self.arg_namespace.start_yearmonth:
                err_text_date_prior_to_dataset = f"""
                        ERROR on -s {start_yearmonth}      ERROR
                            * bad year-month
                                - no data available for dates
                                prior to 2000y
                """
                err_text_date_is_in_future = f"""
                        ERROR on -s {start_yearmonth}      ERROR
                            * year-month cannot be in the future.
                """
                if int(start_yearmonth) < 200000:
                    raise ValueError(err_text_date_prior_to_dataset)
                if int(start_yearmonth) > int(  # noqa:BLK100
                    get_current_datemonth_gmt_minus5()
                ):
                    raise ValueError(err_text_date_is_in_future)
        except ValueError as err:
            raise SystemExit from err

    def _check_end_yearmonth_in_range(self) -> None:
        """Check -e is not earlier than 200001 or later than current year-month.

        Raises:
            SystemExit: Exit on error
        """
        try:
            if end_yearmonth := self.arg_namespace.end_yearmonth:
                err_text_date_prior_to_dataset = f"""
                        ERROR on -e {end_yearmonth}     ERROR
                            * bad year-month
                                - no data available for dates
                                prior to 2000y
                """
                err_text_date_is_in_future = f"""
                        ERROR on -e {end_yearmonth}     ERROR
                            * year-month cannot be in the future.
                """
                if int(end_yearmonth) < 200000:
                    raise ValueError(err_text_date_prior_to_dataset)
                if int(end_yearmonth) > int(  # noqa:BLK100
                    get_current_datemonth_gmt_minus5()
                ):
                    raise ValueError(err_text_date_is_in_future)
        except ValueError as err:
            raise SystemExit from err

    def _check_start_lessthan_end(self) -> None:
        """Validate -e is not a year-month earlier than -s.

        Raises:
            SystemExit: Exit on error
        """
        try:
            if (start_yearmonth := self.arg_namespace.start_yearmonth) and (
                end_yearmonth := self.arg_namespace.end_yearmonth
            ):

                err_text_start_date_after_end_date = f"""
                    ERROR on -s {start_yearmonth} -e {end_yearmonth}    ERROR
                        * logic error: end year-month is before start year-month.
                    """
                if int(start_yearmonth) > int(end_yearmonth):
                    raise ValueError(err_text_start_date_after_end_date)
        except ValueError as err:
            raise SystemExit from err

    def _validate_yearmonth_format(self, yearmonth: str) -> str | Any:
        # pylint: disable=anomalous-backslash-in-string
        """Validate initial user input.

        checks year-month format for:
            -  0000
            -  0000-00
            - "0000 00"
            -  0000_00
            -  0000.00
            -  0000:00
            -  000000
            -  start
            -  now

        # noqa: DAR402

        Args:
            yearmonth (str): YYYYMM

        Raises:
            ValueError: invalid yearmonth format
            SystemExit: Exit on error

        Returns:
            str | Any: str("YYYYMM") or "now" or "start"
        """
        try:
            if (
                re.match("^\d{4}[-_.: ]\d{2}$", yearmonth)  # noqa:W605
                or re.match("^\d{6}$", yearmonth)  # noqa:W605
                or re.match("^\d{4}$", yearmonth)  # noqa:W605
                or str.lower(yearmonth) == "now"
                or str.lower(yearmonth) == "start"
                or not yearmonth
            ):
                return replace_date_punct(yearmonth)
            err_text_bad_yearmonth_format = f"""
                        ERROR on {yearmonth}    ERROR
                            * invalid yearmonth format
            """

            raise ValueError(err_text_bad_yearmonth_format)  # noqa:TC301
        except ValueError as err:
            raise SystemExit from err

    def _set_args(self) -> None:  # noqa:CFQ001
        # pylint: disable=unnecessary-lambda
        """Config CLI arguments and default values."""
        self.add_argument(
            "-V",
            "--validate_urls",
            action="store_true",
            help="Check generated list of URLs as valid download locations",
        )
        self.add_argument(
            "-A",
            "--available_remote_data",
            action="store_true",
            help="list data retrievable from histdata.com",
        )
        self.add_argument(
            "-U",
            "--update_remote_data",
            action="store_true",
            help="update list of data retrievable from histdata.com",
        )
        self.add_argument(
            "--by",
            type=str,
            help=(
                "With -A, -U, to sort --by"
                " [pair_asc, pair_dsc, start_asc, start_dsc]"
            ),
        )
        self.add_argument(
            "-D",
            "--download_data_archives",
            action="store_true",
            help=(
                "download specified pairs/formats/timeframe and"  # noqa:BLK100
                " create data files"
            ),
        )
        self.add_argument(
            "-X",
            "--extract_csvs",
            action="store_true",
            help=(
                "histdata.com delivers zip files."  # noqa:BLK100
                " Use the -X flag to extract them."
            ),
        )
        self.add_argument(
            "-I",
            "--import_to_influxdb",
            action="store_true",
            help=(
                "import data to influxdb instance."  # noqa:BLK100
                " Use influxdb.yaml to configure."
            ),
        )
        self.add_argument(
            "-c",
            "--cpu_utilization",
            type=str,
            help=(
                '"low", "medium", "high". High uses all'
                " available CPUs OR integer percent 1-200"
            ),  # noqa: E501
        )
        self.add_argument(
            "-p",
            "--pairs",
            nargs="+",
            type=str,
            choices=Pairs.list_keys(),
            help="space separated currency pairs. e.g. -p eurusd usdjpy ...",
            metavar="PAIR",
        )
        self.add_argument(
            "-f",
            "--formats",
            nargs="+",
            type=str,
            choices=Format.list_values(),
            help=(
                "space separated formats. -f "  # noqa:BLK100
                "metatrader ascii ninjatrader metastock"
            ),
            metavar="FORMAT",
        )
        self.add_argument(
            "-t",
            "--timeframes",
            nargs="+",
            type=(
                lambda v: Timeframe(v).name  # type: ignore
            ),  # convert long Timeframe .value to short .key
            choices=Timeframe.list_keys(),
            help=(
                "space separated Timeframes. -t "  # noqa:BLK100
                "tick-data-quotes 1-minute-bar-quotes"
            ),
            metavar="TIMEFRAME",
        )
        self.add_argument(
            "-s",
            "--start_yearmonth",
            type=(lambda v: self._validate_yearmonth_format(v)),
            help=(
                "set a start year and month for data."  # noqa:BLK100
                " e.g. -s 2000-04 or -s 2015-00"
            ),
        )
        self.add_argument(
            "-e",
            "--end_yearmonth",
            type=(lambda v: self._validate_yearmonth_format(v)),
            help=(
                "set a start year and month for data."  # noqa:BLK100
                " e.g. -e 2020-00 or -e 2022-04"
            ),
        )
        self.add_argument(
            "-b",
            "--batch_size",
            type=int,
            help="(integer) influxdb write_api batch size. defaults to 5000",
        )
        self.add_argument(
            "-d",
            "--delete_after_influx",
            action="store_true",
            help="delete data files after upload to influxdb",
        )
        self.add_argument(
            "--data-directory",
            type=str,
            help='Directory Used to save data. default is "./data/"',
        )
        self.add_argument(
            "--version",
            action="store_true",
            help="return current version of histdatacom.",
        )

    def _sanitize_input(self) -> None:
        """Clean user-input before run."""
        # prevent running from cli with no arguments
        if len(sys.argv) == 1 and not self.arg_namespace.from_api:
            self.print_help(sys.stdout)
            sys.exit(0)

        self._adjust_for_repo_data_request()

        if "histdatacom" not in sys.argv[0] and self.arg_namespace.from_api:
            args = self._clean_from_api_args()
            self._false_from_api_if_behavior_flag()
            self.parse_args(args, namespace=self.arg_namespace)
        else:
            # Get the args from sys.argv
            self.parse_args(namespace=self.arg_namespace)

        self._check_datetime_input()
        self._check_for_ascii_if_influx()
        self._check_for_ascii_if_api()
        get_pool_cpu_count(self.arg_namespace.cpu_utilization)

    def __call__(self) -> Options:
        """Collect and process settings from CLI or API.

        Returns:
            Options: arg_namespace for config.ARGS
        """
        self._set_args()
        self._sanitize_input()
        return self.arg_namespace



"""Api functions for histdatacom.

Raises:
    ValueError: On failed jay creation
    SystemExit: On failed jay creation

Returns:
    "Frame" | "DataFrame" | "Table":
        - (Frame) if options.api_return_type = "datatable"
        - (DataFrame) if options.api_return_type = "pandas"
        - (Table) if options.api_return_type = "arrow"
"""
from __future__ import annotations

import itertools
import sys  # sourcery skip
from pathlib import Path
from typing import TYPE_CHECKING, Tuple

import datatable as dt  # noqa:I900
from datatable import Frame, f, update  # noqa:I900
from rich.progress import BarColumn, Progress, TextColumn, TimeElapsedColumn

from histdatacom import config
from histdatacom.concurrency import ProcessPool, get_pool_cpu_count
from histdatacom.scraper.scraper import Scraper
from histdatacom.utils import check_installed_module

if TYPE_CHECKING:
    from datatable import FExpr  # noqa:I900
    from pandas.core.frame import DataFrame
    from pyarrow import Table

    from histdatacom.records import Record, Records

dt.options.progress.enabled = False


class Api:  # noqa:H601
    """Api functions for histdatacom.

    Raises:
        ValueError: On failed jay creation
        SystemExit: On failed jay creation

    Returns:
        "Frame" | "DataFrame" | "Table":
            - (Frame) if options.api_return_type = "datatable"
            - (DataFrame) if options.api_return_type = "pandas"
            - (Table) if options.api_return_type = "arrow"
    """

    @classmethod
    def _create_jay(cls, record: "Record", args: dict) -> None:
        """Create .jay file based on single record's data.

        creates a datatable file, saves it in dt's native jay format
        using and updating relevant information in a Record of work.

        Args:
            record (Record): a histdatacom.records.Record
            args (dict): args received from argparse
        """
        zip_path = Path(record.data_dir, record.zip_filename)
        csv_path = Path(record.data_dir, record.csv_filename)

        if zip_path.exists():
            file_data = cls._import_file_to_datatable(record, zip_path)
        elif csv_path.exists():
            file_data = cls._import_file_to_datatable(record, csv_path)

        record.jay_filename = ".data"
        jay_path = record.data_dir + record.jay_filename
        cls._export_datatable_to_jay(file_data, jay_path)

        record.jay_line_count = file_data.nrows
        record.jay_start = str(
            cls._extract_single_value_from_frame(file_data, 0, "datetime")
        )
        record.jay_end = str(
            cls._extract_single_value_from_frame(
                file_data, file_data.nrows - 1, "datetime"
            )
        )
        record.write_memento_file(base_dir=args["default_download_dir"])

    @classmethod
    def test_for_jay_or_create(cls, record: Record, args: dict) -> None:
        """Test for record's jay file. if it doesn't exist, create it.

           a helper method to ensure the existence of a Record's jay file
           prior to further processing from the API or Influx classes

        Args:
            record (Record): a histdatacom.records.Record
            args (dict): args received from argparse
        """
        if str.lower(  # noqa:BLK001
            record.data_format
        ) == "ascii" and record.data_timeframe in [
            "T",
            "M1",
        ]:
            jay_path = Path(record.data_dir, ".data")
            if not jay_path.exists():
                if "CSV" not in record.status:
                    Scraper.get_zip_file(record)
                cls._create_jay(record, args)

    @classmethod
    def _validate_jay(
        cls,
        record: Record,
        args: dict,
        records_current: Records,
        records_next: Records,
    ) -> None:
        """Validate Jay prior to possible merge operation.

           A Wrapper to be passed to an individual process within the process
           pool to test for or create a datatable jay file based on a Record of
           Work's information.  Receives a unit of work from the pool, performs
           validation, readies the Record for further processing, and marks the
           current work as complete.

        Args:
            record (Record): a Histdatacom.records.Record
            args (dict): arguments received from argparse
            records_current (Records):
                Current Work Records Queue
            records_next (Records):
                Records Queue for Further Work

        Raises:
            Exception: Unknown Exception
        """
        try:
            cls.test_for_jay_or_create(record, args)
            records_next.put(record)
        except Exception:
            print("Unexpected error:", sys.exc_info())  # noqa:T201
            record.delete_momento_file()
            raise
        finally:
            records_current.task_done()

    @classmethod
    def _extract_single_value_from_frame(
        cls, frame: "Frame", row: int, column: str
    ) -> int:
        """Extract Single Value from datatable Frame.

        Args:
            frame (Frame): datatable.Frame
            row (int): datatable frame row
            column (str): datatable frame column

        Returns:
            int: datatable time64
        """
        return int(frame[row, column])

    @classmethod
    def _import_file_to_datatable(cls, record: Record, zip_path: Path) -> Frame:
        """Import file as datatable Frame and convert timestamp to UTC.

        # noqa: DAR402

        Args:
            record (Record): a record from the work queue.
            zip_path (Path): path to record's zip.

        Raises:
            ValueError: Error on null value.
            SystemExit: Exit on Error.


        Returns:
            Frame: datatable.Frame
        """
        # pylint: disable=expression-not-assigned
        try:
            match record.data_timeframe:
                case "M1":
                    frame_data: Frame = cls._import_frame_with_headers(
                        record.data_timeframe, zip_path
                    )

                    ascii_m1_etc_ms_timestamp = cls._strptime_fexpr_for_frame(
                        record.data_timeframe
                    )

                    ascii_m1_utc_ms_timestamp = (  # noqa:BLK001
                        cls._adjust_est_timestamp_to_utc(  # noqa:BLK001
                            ascii_m1_etc_ms_timestamp
                        )
                    )

                    # pylint: disable-next=unsubscriptable-object
                    frame_data[:, update(datetime=ascii_m1_utc_ms_timestamp)]
                case "T":
                    frame_data = cls._import_frame_with_headers(
                        record.data_timeframe, zip_path
                    )

                    ascii_t_etc_ms_timestamp = cls._strptime_fexpr_for_frame(
                        record.data_timeframe
                    )

                    ascii_t_utc_ms_timestamp = cls._adjust_est_timestamp_to_utc(
                        ascii_t_etc_ms_timestamp
                    )

                    # pylint: disable-next=unsubscriptable-object
                    frame_data[:, update(datetime=ascii_t_utc_ms_timestamp)]
                case _:
                    raise ValueError

            # pylint: disable-next=unsupported-assignment-operation
            frame_data["vol"] = dt.int32
            return frame_data  # noqa:TC300
        except ValueError as err:
            raise SystemExit from err

    @classmethod
    def _import_frame_with_headers(  # noqa:BLK001
        cls, timeframe: str, zip_path: Path
    ) -> "Frame":
        """Import datatable.Frame with headers for M1 and Tick data.

        Args:
            timeframe (str): M1 or T
            zip_path (Path): Path

        Raises:
            ValueError: when not M1 or T

        Returns:
            Frame: datatable.Frame
        """
        match timeframe:
            case "M1":
                frame = dt.fread(
                    zip_path,
                    header=False,
                    columns=[
                        "datetime",
                        "open",
                        "high",
                        "low",
                        "close",
                        "vol",
                    ],
                    multiple_sources="ignore",
                )
            case "T":
                frame = dt.fread(
                    zip_path,
                    header=False,
                    columns=["datetime", "bid", "ask", "vol"],
                    multiple_sources="ignore",
                )
            case _:
                raise ValueError
        return frame  # noqa:R504

    @classmethod
    def _strptime_fexpr_for_frame(cls, timeframe: str) -> "FExpr":
        """Convert csv datetime to unix timestamp in EST.

           Convert csv datetime to unix timestamp in EST with NO
           DAYLIGHT SAVINGS TIME.

        Args:
            timeframe (str): M1 or T

        Returns:
            FExpr: datatable.FExpr
        """
        match timeframe:
            case "M1":
                ascii_m1_str_splitter = dt.time.ymdt(  # sourcery skip
                    f.datetime[0:4].as_type(int),
                    f.datetime[4:6].as_type(int),
                    f.datetime[6:8].as_type(int),
                    f.datetime[9:11].as_type(int),
                    f.datetime[11:13].as_type(int),
                    f.datetime[13:15].as_type(int),
                )
                temp_time = ascii_m1_str_splitter.as_type(int)
            case "T":
                ascii_t_str_splitter = dt.time.ymdt(
                    f.datetime[0:4].as_type(int),
                    f.datetime[4:6].as_type(int),
                    f.datetime[6:8].as_type(int),
                    f.datetime[9:11].as_type(int),
                    f.datetime[11:13].as_type(int),
                    f.datetime[13:15].as_type(int),
                    10**6 * f.datetime[15:18].as_type(int),
                )
                temp_time = ascii_t_str_splitter.as_type(int)

        return temp_time // 10**6

    @classmethod
    def _adjust_est_timestamp_to_utc(cls, est_fexpr: "FExpr") -> "FExpr":
        """Convert ESTnoDST millisecond timestamp to UTC.

        Args:
            est_fexpr (FExpr): datatable.FExpr

        Returns:
            FExpr: datatable.FExpr
        """
        return est_fexpr + 18000000

    @classmethod
    def _export_datatable_to_jay(  # noqa:BLK001
        cls, data_frame: "Frame", file_path: str
    ) -> None:
        """Export datatable frame to jay.

        Args:
            data_frame (Frame): datatable.Frame
            file_path (str): dest path
        """
        data_path = file_path
        data_frame.to_jay(data_path)

    @classmethod
    def import_jay_data(cls, jay_path: str) -> "Frame":
        """Read jay file in to datatable Frame.

        Args:
            jay_path (str): source path

        Returns:
            Frame: datatable.Frame
        """
        return dt.fread(jay_path)

    def validate_jays(self) -> None:
        """Initialize a process pool to validate jay files."""
        pool = ProcessPool(
            self._validate_jay,
            config.ARGS,
            "Staging",
            "data files...",
            get_pool_cpu_count(config.ARGS["cpu_utilization"]),
        )
        pool(config.CURRENT_QUEUE, config.NEXT_QUEUE)

    def merge_jays(self) -> list | Frame | DataFrame | Table:
        """Merge jays for start_yearmonth and end_yearmonth range.

        Returns:
            list | Frame | DataFrame | Table: _description_
        """
        records_to_merge: list
        pairs: list
        timeframes: list

        records_to_merge, pairs, timeframes = self._dequeue_records_for_merge()
        sets_to_merge: list = self._collate_sets_to_merge(
            records_to_merge, pairs, timeframes
        )

        for tp_set in sets_to_merge:
            self._merge_records(tp_set)

        return (  # noqa:BLK001
            sets_to_merge[0]["data"]  # noqa:BLK001
            if len(sets_to_merge) == 1
            else sets_to_merge
        )

    def _merge_records(self, tp_set_dict: dict) -> None:
        """Sort and Merge records from a timeframe/pair set.

        Args:
            tp_set_dict (dict): {
                                 "timeframe": timeframe,
                                 "pair": pair,
                                 "records": [],
                                 "data": datatable.Frame | None,
                                }
        """
        match tp_set_dict["timeframe"]:
            case "T":
                merged = dt.Frame(names=["datetime", "bid", "ask", "vol"])
            case "M1":
                merged = dt.Frame(
                    names=[
                        "datetime",
                        "open",
                        "high",
                        "low",
                        "close",
                        "vol",
                    ]
                )

        tp_set_dict["records"].sort(key=lambda record: record.jay_start)

        records_count = len(tp_set_dict)
        with Progress(
            TextColumn(text_format="[cyan]Merging records..."),
            BarColumn(),
            "[progress.percentage]{task.percentage:>3.0f}%",
            TimeElapsedColumn(),
        ) as progress:
            progress.add_task("merge", total=records_count)

            for m_record in tp_set_dict["records"]:
                jay_path = m_record.data_dir + m_record.jay_filename
                jay_data = self.import_jay_data(jay_path)
                merged.rbind(jay_data)

            match config.ARGS["api_return_type"]:
                case "datatable":
                    tp_set_dict["data"] = merged
                case "arrow":
                    check_installed_module("arrow", True)
                    tp_set_dict["data"] = merged.to_arrow()
                case "pandas":
                    check_installed_module("pandas", True)
                    tp_set_dict["data"] = merged.to_pandas()

    def _dequeue_records_for_merge(self) -> Tuple[list, list, list]:
        """Empty the queue of relevant records.

        Empty the queue of relevant records, create additional
        lists for each pair and timeframe to appear in the set
        of records.

        Returns:
            Tuple[list, list, list]: records_to_merge, pairs, timeframes
        """
        records_to_merge: list = []
        pairs: list = []
        timeframes: list = []
        while not config.CURRENT_QUEUE.empty():  # type: ignore
            record = config.CURRENT_QUEUE.get()  # type: ignore

            if record is None:
                break

            if (
                record.jay_filename == ".data"
                and Path(record.data_dir, record.jay_filename).exists()
            ):
                pairs.append(record.data_fxpair)
                timeframes.append(record.data_timeframe)
                records_to_merge.append(record)

        return records_to_merge, pairs, timeframes

    def _collate_sets_to_merge(
        self, records_to_merge: list, pairs: list, timeframes: list
    ) -> list:  # noqa:TAE002
        """Organize records into pair/timeframe sets.

        Args:
            records_to_merge (list): list of records
            pairs (list): list of pairs
            timeframes (list): list of timeframes

        Returns:
            list: _description_
        """
        sets_to_merge = []
        for timeframe, pair in itertools.product(set(timeframes), set(pairs)):
            tp_set_dict = {
                "timeframe": timeframe,
                "pair": pair,
                "records": [],
                "data": None,
            }
            for m_record in records_to_merge:
                if (
                    m_record.data_timeframe == timeframe
                    and m_record.data_fxpair == pair
                ):
                    tp_set_dict["records"].append(m_record)
            sets_to_merge.append(tp_set_dict)

        return sets_to_merge



"""Entry point for histdatacom api.

histdatacom(options)

Returns:
    data: returns a data frame or a list of data frames and metadata
"""
from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from histdatacom.fx_enums import Format, Pairs, Timeframe
from histdatacom.options import Options

from . import histdata_com  # noqa:WPS130

if TYPE_CHECKING:
    from datatable import Frame  # noqa:I900
    from pandas import DataFrame
    from pyarrow import Table

__all__ = [
    "Options",
    "Pairs",
    "Timeframe",
    "Format",
]


__version__ = "0.78.1"
__author__ = "David Midlo"


class APICaller(sys.modules[__name__].__class__):  # type: ignore # noqa:H601
    """APICaller. A Masquerade class.

    A class that extends sys.modules[__name__].__class__ (the histdatacom class)
    extends/overwrites with a __call__ method to allow the module to be callable.

    Returns:
        data: returns a data frame or a list of data frames and metadata
    """

    def __call__(  # noqa:BLK001
        self, options: Options
    ) -> "list" | "Frame" | "DataFrame" | "Table":
        """Run histdatacom -h for help.

        Args:
            options (Options): a histdatacom Options object.

        Returns:
            "list" | "Frame" | "DataFrame" | "Table":
                - (list) if called with -A or -U
                - (Frame) if options.api_return_type = "datatable"
                - (DataFrame) if options.api_return_type = "pandas"
                - (Table) if options.api_return_type = "arrow"
        """
        return histdata_com.main(options)


sys.modules[__name__].__class__ = APICaller



"""Provide multi-threading and multi-processes facilities.

Raises:
    ValueError: -c cpu must be str:
                    low, medium, or high. or integer percent 1-200
    SystemExit: exit on error
"""
# pylint: disable=redefined-outer-name
from __future__ import annotations

from concurrent.futures import (
    Future,
    ProcessPoolExecutor,
    ThreadPoolExecutor,
    as_completed,
)
from math import ceil
from multiprocessing import Queue, cpu_count, managers
from typing import TYPE_CHECKING, Callable, Optional, Type

from rich.progress import (
    BarColumn,
    Progress,
    TaskID,
    TextColumn,
    TimeElapsedColumn,
)

from histdatacom import config
from histdatacom.records import Records

if TYPE_CHECKING:
    from datatable import Frame  # noqa:I900
    from pandas import DataFrame
    from pyarrow import Table

    from histdatacom.histdata_com import _HistDataCom
    from histdatacom.options import Options


class ThreadPool:
    """Standardize thread pool execution for histdatacom."""

    def __init__(
        self,
        exec_func: Callable,
        args: dict,
        progress_pre_text: str,
        progress_post_text: str,
        cpu_count: int,
    ) -> None:
        """Initialize attributes for thread pool.

        Args:
            exec_func (Callable): function to be executed by pool.
            args (dict): global args from config.ARGS
            progress_pre_text (str): display for rich.Progress
            progress_post_text (str): display for rich.Progress
            cpu_count (int): CPU count to use for pool.
        """
        self.exec_func = exec_func
        self.args = args
        self.progress_pre_text = progress_pre_text
        self.progress_post_text = progress_post_text
        self.cpu_count = cpu_count

    def __call__(
        self,
        records_current: Optional[Records],
        records_next: Optional[Records],
    ) -> None:
        """Execute Thread pool with rich.Progress bar.

        Args:
            records_current (Optional[Records]): from config.CURRENT_QUEUE
            records_next (Optional[Records]): from config.NEXT_QUEUE
        """
        records_count = records_current.qsize()  # type: ignore
        with Progress(
            TextColumn(
                text_format=(
                    f"[cyan]{self.progress_pre_text} "
                    f"{records_count} "
                    f"{self.progress_post_text}."
                )
            ),
            BarColumn(),
            "[progress.percentage]{task.percentage:>3.0f}%",
            TimeElapsedColumn(),
        ) as progress:
            task_id = progress.add_task("Validating URLs", total=records_count)

            with ThreadPoolExecutor(
                max_workers=self.cpu_count,
                initializer=_init_counters,
                initargs=(records_current, records_next, self.args.copy()),
            ) as executor:
                futures = []

                while not records_current.empty():  # type: ignore
                    record = records_current.get()  # type: ignore

                    if record is None:
                        return

                    future = executor.submit(self.exec_func, record, self.args)
                    progress.advance(task_id, 0.25)
                    futures.append(future)

                for future in as_completed(futures):
                    _complete_future(progress, task_id, futures, future)

        records_current.join()  # type: ignore
        records_next.dump_to_queue(records_current)  # type: ignore


class ProcessPool:
    """Standardize process pool execution for histdatacom."""

    def __init__(  # noqa:CFQ002
        self,
        exec_func: Callable,
        args: dict,
        progress_pre_text: str,
        progress_post_text: str,
        cpu_count: int,
        join: bool = True,
        dump: bool = True,
    ) -> None:
        """Initialize attributes for process pool.

        Args:
            exec_func (Callable): function to be executed by pool.
            args (dict): global args from config.ARGS
            progress_pre_text (str): display for rich.Progress
            progress_post_text (str): display for rich.Progress
            cpu_count (int): CPU count to use for pool.
            join (bool): disable join waits. Defaults to True.
            dump (bool): enable queue dumps. Defaults to True.
        """
        self.exec_func = exec_func
        self.args = args
        self.progress_pre_text = progress_pre_text
        self.progress_post_text = progress_post_text
        self.cpu_count = cpu_count
        self.join = join
        self.dump = dump

    def __call__(  # noqa:CCR001
        self,
        records_current: Optional[Records],
        records_next: Optional[Records],
        influx_chunks_queue: Optional[Queue] = None,
    ) -> None:
        """Execute Process pool with rich.Progress bar.

        Args:
            records_current (Optional[Records]): _description_
            records_next (Optional[Records]): _description_
            influx_chunks_queue (Optional[Queue], optional):
                                    used for RxPY queue. Defaults to None.
        """
        records_count = records_current.qsize()  # type: ignore
        with Progress(
            TextColumn(
                text_format=(
                    f"[cyan]{self.progress_pre_text} "
                    f"{records_count} "
                    f"{self.progress_post_text}."
                )
            ),
            BarColumn(),
            "[progress.percentage]{task.percentage:>3.0f}%",
            TimeElapsedColumn(),
        ) as progress:
            task_id = progress.add_task("Validating URLs", total=records_count)

            with ProcessPoolExecutor(
                max_workers=self.cpu_count,
                initializer=_init_counters,
                initargs=(
                    records_current,
                    records_next,
                    self.args.copy(),
                    influx_chunks_queue,
                ),
            ) as executor:
                futures = []

                while not records_current.empty():  # type: ignore
                    record = records_current.get()  # type: ignore

                    if record is None:
                        return

                    if influx_chunks_queue is None:
                        future = executor.submit(
                            self.exec_func,
                            record,
                            self.args,
                            records_current,
                            records_next,
                        )
                    else:
                        future = executor.submit(
                            self.exec_func,
                            record,
                            self.args,
                            records_current,
                            records_next,
                            influx_chunks_queue,
                        )

                    progress.advance(task_id, 0.25)
                    futures.append(future)

                for future in as_completed(futures):
                    _complete_future(progress, task_id, futures, future)

        if self.join:
            records_current.join()  # type: ignore

        if self.dump:
            records_next.dump_to_queue(records_current)  # type: ignore


def get_pool_cpu_count(count: str | int | None = None) -> int:  # noqa:CCR001
    """Set cpu_count.  Adjusted by -c config.ARGS["cpu_utilization"].

    Args:
        count (str | int | None, optional):
                Defaults to multiprocessing.cpu_count().

    # noqa: DAR402

    Raises:
        ValueError: malformed command, bad -c cpu_count
        SystemExit: exit on error.

    Returns:
        int: cpu_count for thread & process pool.
    """
    try:
        real_vcpu_count = cpu_count()

        if count is None:
            count = real_vcpu_count
        else:
            err_text_cpu_level_err = f"""
                    ERROR on -c {count}  ERROR
                        * Malformed command:
                            - -c cpu must be str:
                                low, medium, or high. or integer percent 1-200
            """
            count = str(count)
            match count:
                case "low":
                    count = ceil(real_vcpu_count / 2.5)
                case "medium":
                    count = ceil(real_vcpu_count / 1.5)
                case "high":
                    count = real_vcpu_count
                case _:
                    if count.isnumeric() and 1 <= int(count) <= 200:
                        count = ceil(real_vcpu_count * (int(count) / 100))
                    else:
                        raise ValueError(err_text_cpu_level_err)

        return count - 1 if count > 2 else ceil(count / 2)  # noqa:TC300
    except ValueError as err:
        raise SystemExit from err


class QueueManager:
    """Configure SyncManager with Queues and Callable: histdatacom."""

    # pylint: disable=no-member
    def __init__(self, options: Options):
        """Initialize SyncManager and register custom queue class 'Records'.

        Args:
            options (Options): a histdatacom.options Options object.
        """
        self.options = options
        config.QUEUE_MANAGER = managers.SyncManager()
        config.QUEUE_MANAGER.register(  # noqa:BLK100
            "Records", Records
        )  # pylint: disable=no-member

    # pylint: disable-next=inconsistent-return-statements
    def __call__(  # type: ignore
        self, runner_: Type[_HistDataCom]
    ) -> list | dict | Frame | DataFrame | Table:
        """Configure global queues and execute.

        Args:
            runner_ (Type[_HistDataCom]): _description_

        Returns:
            list | dict | Frame | DataFrame | Table: _description_
        """
        # pylint: disable-next=consider-using-with
        config.QUEUE_MANAGER.start()  # type: ignore

        config.CURRENT_QUEUE = config.QUEUE_MANAGER.Records()  # type: ignore
        config.NEXT_QUEUE = config.QUEUE_MANAGER.Records()  # type: ignore
        config.INFLUX_CHUNKS_QUEUE = config.QUEUE_MANAGER.Queue()  # type: ignore

        histdatacom_runner = runner_(self.options)

        if self.options.from_api:
            return histdatacom_runner.run()

        histdatacom_runner.run()  # noqa:R503


def _init_counters(
    # pylint: disable=unused-argument,unused-variable
    records_current_: Records,
    records_next_: Records,
    args_: dict,
    influx_chunks_queue_: Queue | None = None,
) -> None:
    """Initialize pool with access to these global variables.

    Args:
        records_current_ (Records): config.CURRENT_QUEUE
        records_next_ (Records): config.NEXT_QUEUE
        args_ (dict): config.ARGS
        influx_chunks_queue_ (Queue | None, optional): config.INFLUX_CHUNKS_QUEUE
    """
    # pylint: disable=global-variable-undefined
    args = args_  # noqa:F841

    if influx_chunks_queue_ is not None:
        global INFLUX_CHUNKS_QUEUE  # noqa:WPS100
        INFLUX_CHUNKS_QUEUE = influx_chunks_queue_  # type: ignore


def _complete_future(
    progress: Progress, task_id: TaskID, futures: list, future: Future
) -> None:
    """Finalize future and rich.Progress task.

    Args:
        progress (Progress): progress bar instance.
        task_id (TaskID): progress instance task id.
        futures (list): list of futures.
        future (Future): future from pool.
    """
    progress.advance(task_id, 0.75)
    futures.remove(future)
    del future  # noqa:WPS100



"""Configure Enums used by histdatacom."""
# pylint: disable=invalid-name
from enum import Enum

from influxdb_client import WritePrecision

# Majors 7
# eurusd usdjpy gbpusd usdcad usdchf audusd nzdusd

# Minors 7
# eurgbp euraud gbpchf audnzd audcad audchf gbpaud usdmxn

# Crosses 10
# eurchf eurcad eurnzd eurjpy gbpjpy chfjpy cadjpy
# audjpy nzdjpy gbpcad nzdcad sgdjpy gbpnzd cadchf

# Exotics 7
# eurtry usdtry usdsek usdnok usddkk usdzar usdhkd
# usdsgd eurpln eurhuf nzdchf usdhuf usdpln eurczk
# eursek usdczk zarjpy eurdkk eurnok usddkk

# metals-commodities
# xauusd xauaud xauchf bcousd wtiusd xaueur xagusd xaugbp

# indices
# grxeur auxaud frxeur hkxhkd spxusd jpxjpy udxusd
# nsxusd ukxgbp etxeur


class Pairs(Enum):  # noqa:H601
    """Enumerate list of public pairs published on histdata.com.

    Args:
        Enum (Enum): Parent Class

    Returns:
        (set): set of keys or values with:
                Pairs.list_keys()
                Pairs.list_values()
    """

    eurusd = "EUR_USD"
    eurchf = "EUR_CHF"
    eurgbp = "EUR_GBP"
    eurjpy = "EUR_JPY"
    euraud = "EUR_AUD"
    usdcad = "USD_CAD"
    usdchf = "USD_CHF"
    usdjpy = "USD_JPY"
    usdmxn = "USD_MXN"
    gbpchf = "GBP_CHF"
    gbpjpy = "GBP_JPY"
    gbpusd = "GBP_USD"
    audjpy = "AUD_JPY"
    audusd = "AUD_USD"
    chfjpy = "CHF_JPY"
    nzdjpy = "NZD_JPY"
    nzdusd = "NZD_USD"
    xauusd = "XAU_USD"
    eurcad = "EUR_CAD"
    audcad = "AUD_CAD"
    cadjpy = "CAD_JPY"
    eurnzd = "EUR_NZD"
    grxeur = "GRX_EUR"
    nzdcad = "NZD_CAD"
    sgdjpy = "SGD_JPY"
    usdhkd = "USD_HKD"
    usdnok = "USD_NOK"
    usdtry = "USD_TRY"
    xauaud = "XAU_AUD"
    audchf = "AUD_CHF"
    auxaud = "AUX_AUD"
    eurhuf = "EUR_HUF"
    eurpln = "EUR_PLN"
    frxeur = "FRX_EUR"
    hkxhkd = "HKX_HKD"
    nzdchf = "NZD_CHF"
    spxusd = "SPX_USD"
    usdhuf = "USD_HUF"
    usdpln = "USD_PLN"
    usdzar = "USD_ZAR"
    xauchf = "XAU_CHF"
    zarjpy = "ZAR_JPY"
    bcousd = "BCO_USD"
    etxeur = "ETX_EUR"
    eurczk = "EUR_CZK"
    eursek = "EUR_SEK"
    gbpaud = "GBP_AUD"
    gbpnzd = "GBP_NZD"
    jpxjpy = "JPX_JPY"
    udxusd = "UDX_USD"
    usdczk = "USD_CZK"
    usdsek = "USD_SEK"
    wtiusd = "WTI_USD"
    xaueur = "XAU_EUR"
    audnzd = "AUD_NZD"
    cadchf = "CAD_CHF"
    eurdkk = "EUR_DKK"
    eurnok = "EUR_NOK"
    eurtry = "EUR_TRY"
    gbpcad = "GBP_CAD"
    nsxusd = "NSX_USD"
    ukxgbp = "UKX_GBP"
    usddkk = "USD_DKK"
    usdsgd = "USD_SGD"
    xagusd = "XAG_USD"
    xaugbp = "XAU_GBP"

    @classmethod
    def list_keys(cls) -> set:
        """Return a set containing the keys of the Pairs enum.

        Returns:
            Pairs (set): set of str(keys)
        """
        return set(cls.__members__.keys())

    @classmethod
    def list_values(cls) -> set:
        """Return a set containing the values of the Pairs enum.

        Returns:
            Pairs (set): set of str(values)
        """
        return {member.value for _, member in cls.__members__.items()}


class Format(Enum):  # noqa:H601
    """Enumerate list of public formats published on histdata.com.

    Args:
        Enum (Enum): Parent Class

    Returns:
        (set): set of keys or values with:
                Format.list_keys()
                Format.list_values()
    """

    MT = "metatrader"
    NT = "ninjatrader"
    MS = "metastock"
    ASCII = "ascii"
    XLSX = "excel"

    @classmethod
    def list_keys(cls) -> set:
        """Return a set containing the keys of the Format enum.

        Returns:
            Pairs (set): set of str(keys)
        """
        return set(cls.__members__.keys())

    @classmethod
    def list_values(cls) -> set:
        """Return a set containing the values of the Format enum.

        Returns:
            Format (set): set of str(values)
        """
        return {member.value for _, member in cls.__members__.items()}


class Timeframe(Enum):  # noqa:H601
    """Enumerate list of public Timeframes published on histdata.com.

    Args:
        Enum (Enum): Parent Class

    Returns:
        (set): set of keys or values with:
                Timeframe.list_keys()
                Timeframe.list_values()
    """

    M1 = "1-minute-bar-quotes"
    T = "tick-data-quotes"  # noqa:WPS121
    T_LAST = "tick-last-quotes"
    T_BID = "tick-bid-quotes"
    T_ASK = "tick-ask-quotes"

    @classmethod
    def list_keys(cls) -> set:
        """Return a set containing the keys of the Timeframe enum.

        Returns:
            Pairs (set): set of str(keys)
        """
        return set(cls.__members__.keys())

    @classmethod
    def list_values(cls) -> set:
        """Return a set containing the values of the Timeframe enum.

        Returns:
            Format (set): set of str(values)
        """
        return {member.value for _, member in cls.__members__.items()}

    @classmethod
    def convert_to_values(cls, timeframe_set: set) -> set:
        """Convert a mixed set of timeframe keys and values to values only.

        Args:
            timeframe_set (set): mixed set of keys and values.

        Returns:
            set: timeframe values.
        """
        timeframes = timeframe_set
        if intersect := timeframe_set & Timeframe.list_keys():
            timeframes = timeframe_set | {
                Timeframe[timeframe].value
                for timeframe in intersect
                if timeframe in Timeframe.list_keys()
            }
        return timeframes & Timeframe.list_values()


class TimeFormat(Enum):  # noqa:H601
    """Enumerate list of public Timeformats for strptime histdata.com.

        from histdata.com csv sources.

    Args:
        Enum (Enum): Parent Class

    Returns:
        (set): set of keys or values with:
                TimeFormat.list_keys()
                TimeFormat.list_values()
    """

    MT_M1 = "%Y.%m.%d %H:%M"
    ASCII_M1 = "%Y%m%d %H%M%S"
    ASCII_T = "%Y%m%d %H%M%S%f"
    NT_M1 = "%Y%m%d %H%M%S"  # noqa:PIE796
    NT_T_LAST = "%Y%m%d %H%M%S"  # noqa:PIE796
    NT_T_BID = "%Y%m%d %H%M%S"  # noqa:PIE796
    NT_T_ASK = "%Y%m%d %H%M%S"  # noqa:PIE796
    MS_M1 = "%Y%m%d%H%M"

    @classmethod
    def list_keys(cls) -> set:
        """Return a set containing the keys of the TimeFormat enum.

        Returns:
            Pairs (set): set of str(keys)
        """
        return set(cls.__members__.keys())

    @classmethod
    def list_values(cls) -> set:
        """Return a set containing the values of the TimeFormat enum.

        Returns:
            Format (set): set of str(values)
        """
        return {member.value for _, member in cls.__members__.items()}


class TimePrecision(Enum):  # noqa:H601
    """Enumerate list of public TimePrecisions for influxdb writer.

    Args:
        Enum (Enum): Parent Class

    Returns:
        (set): set of keys or values with:
                TimePrecision.list_keys()
                TimePrecision.list_values()
    """

    MT_M1 = WritePrecision.S
    ASCII_M1 = WritePrecision.S
    ASCII_T = WritePrecision.MS
    NT_M1 = WritePrecision.S
    NT_T_LAST = WritePrecision.S
    NT_T_BID = WritePrecision.S
    NT_T_ASK = WritePrecision.S
    MS_M1 = WritePrecision.S

    @classmethod
    def list_keys(cls) -> set:
        """Return a set containing the keys of the TimePrecision enum.

        Returns:
            Pairs (set): set of str(keys)
        """
        return set(cls.__members__.keys())

    @classmethod
    def list_values(cls) -> set:
        """Return a set containing the values of the TimePrecision enum.

        Returns:
            Format (set): set of str(values)
        """
        return {member.value for _, member in cls.__members__.items()}


def get_valid_format_timeframes(file_format: str) -> list:
    """Return a list of format and timeframe combinations.

    Args:
        file_format (str): _description_

    Returns:
        list: _description_
    """
    timeframes = []

    match file_format:
        case "metatrader":
            timeframes.extend(["M1"])
        case "ninjatrader":
            timeframes.extend(["M1", "T_LAST", "T_BID", "T_ASK"])
        case "metastock":
            timeframes.extend(["M1"])
        case "ascii":
            timeframes.extend(["M1", "T"])
        case "excel":
            timeframes.extend(["M1"])

    return timeframes



"""Options object for histdatacom."""

from histdatacom.fx_enums import Pairs
from histdatacom.fx_enums import Format
from histdatacom.fx_enums import Timeframe


class Options:
    """An intra-class DTO for Default Arguments for _HistDataCom class."""

    # argparse uses a thin class to create a namespace for cli/shell arguments
    # to live in normally argparse.ArgumentParser.parse_args(namespace=...)
    # creates this namespace and writes user's cli args to it.  Preemptively
    # creating here to hold default args; if the user enters args in the shell,
    # these values will be respectively overwritten
    def __init__(self) -> None:
        """Initialize attributes with default values."""
        self.version: bool = False
        self.available_remote_data: bool = False
        self.update_remote_data: bool = False
        self.by: str = "pair_asc"  # pylint: disable=invalid-name
        self.validate_urls: bool = False
        self.download_data_archives: bool = False
        self.extract_csvs: bool = False
        self.import_to_influxdb: bool = False
        self.pairs: set = Pairs.list_keys()
        self.formats: set = Format.list_values()
        self.timeframes: set = Timeframe.list_keys()
        self.start_yearmonth: str | None = ""
        self.end_yearmonth: str | None = ""
        self.data_directory: str = "data"
        self.from_api: bool = False
        self.api_return_type: str | None = None
        self.cpu_utilization: str = "medium"
        self.batch_size: str = "5000"
        self.delete_after_influx: bool = False
        self.zip_persist: bool = False



"""Configure global attributes."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from multiprocessing.managers import SyncManager

    from histdatacom.records import Records

ARGS: dict = {}
REPO_DATA: dict = {}
REPO_DATA_FILE_EXISTS: bool = False
FILTER_PAIRS: set | None = None

QUEUE_MANAGER: SyncManager | None = None

CURRENT_QUEUE: Records | None = None
NEXT_QUEUE: Records | None = None
INFLUX_CHUNKS_QUEUE: None = None

REQUESTS_TIMEOUT: int = 10

POST_HEADERS: dict = {
    "Host": "www.histdata.com",
    "Connection": "keep-alive",
    "Content-Length": "101",
    "Cache-Control": "max-age=0",
    "Origin": "http://www.histdata.com",
    "Upgrade-Insecure-Requests": "1",
    "DNT": "1",
    "Content-Type": "application/x-www-form-urlencoded",
    "Accept": (
        "text/html,application/xhtml+xml,application/xml;"
        "q=0.9,image/avif,image/webp,image/apng,*/*;"
        "q=0.8,application/signed-exchange;"
        "v=b3;"
        "q=0.9"
    ),
    "Referer": "",
    "Accept-Encoding": "gzip, deflate",
    "Accept-Language": "en-US,en;q=0.9",
}



"""Check for or generate a list date ranges for instruments available.

Raises:
    SystemExit: Raised when called from CLI

Returns:
    dict: _description_
"""

import contextlib
import json
import ssl
from pathlib import Path
from ssl import SSLCertVerificationError
from typing import TYPE_CHECKING
from urllib.error import URLError
from urllib.request import urlopen

import certifi
from rich import print  # pylint: disable=redefined-builtin
from rich import box
from rich.table import Table

from histdatacom import config
from histdatacom.scraper.scraper import Scraper
from histdatacom.utils import (
    force_datemonth_if_only_year,
    create_full_path,
    hash_dict,
    get_now_utc_timestamp,
    get_year_from_datemonth,
    get_month_from_datemonth,
)

if TYPE_CHECKING:
    from histdatacom.records import Record


class Repo:  # noqa: H601
    """Creates, updates, or retrieves date ranges for histdata.com pairs.

    Attributes:
        repo_url: remote url of pre-generated repository information.
        repo_local_path: local file path for repo data.

    """

    def __init__(self) -> None:
        """Initialize repo class with remote url."""
        self.repo_url = (
            "https://raw.githubusercontent.com/dmidlo/"
            "histdata.com-tools/main/data/.repo"
        )
        self.repo_local_path = Path(  # noqa:BLK100
            config.ARGS["default_download_dir"], ".repo"
        )

    @staticmethod
    def check_if_queue_is_needed() -> bool:
        # pylint: disable=line-too-long
        """Conditions to populate queue for renewing repo data.

        Returns:
            bool: conditonal filter
        """
        return bool(
            (  # noqa:BLK100
                not config.REPO_DATA_FILE_EXISTS
                and config.ARGS["available_remote_data"]
            )
            or config.ARGS["update_remote_data"]
            or config.FILTER_PAIRS
        )

    @staticmethod
    def check_for_repo_action() -> bool:
        """Check to see if -A or -U has been used.

        Returns:
            bool: general truth case for repo actions
        """
        return bool(
            config.ARGS["available_remote_data"]  # noqa:BLK100
            or config.ARGS["update_remote_data"]
        )

    @staticmethod
    def set_repo_datum(record: "Record") -> None:
        """Create and sort individual date ranges for repo.

        Args:
            record (Record): a single downloadable record
                             of pair, year, and month
        """
        datemonth: str = force_datemonth_if_only_year(  # noqa:BLK100
            record.data_datemonth
        )
        pair = record.data_fxpair.lower()

        if pair not in config.REPO_DATA:
            config.REPO_DATA[pair] = {"start": datemonth, "end": datemonth}
        else:
            if int(datemonth) < int(config.REPO_DATA[pair]["start"]):
                config.REPO_DATA[pair]["start"] = datemonth
            if int(datemonth) > int(config.REPO_DATA[pair]["end"]):
                config.REPO_DATA[pair]["end"] = datemonth

    def test_for_repo_data_file(self) -> bool:
        """Test for repo data file and update global boolean.

        Returns:
            bool: existence of repo data file.
        """
        if self.repo_local_path.exists():
            config.REPO_DATA_FILE_EXISTS = True
            return True
        config.REPO_DATA_FILE_EXISTS = False
        return False

    def read_repo_data_file(self) -> None:
        """Read local file repo data. Append/update global working repo data."""
        with self.repo_local_path.open(
            "r", encoding="UTF-8"
        ) as json_read, contextlib.suppress(Exception):
            while True:
                config.REPO_DATA.update(json.load(json_read))

    def update_repo_from_github(self) -> None:
        """Fetch remote repo data.

        Diffs hash of current repo data with remote. If hash is different and
        remote timestamp is more recent than local file timestamp, overwrite
        local data with remote data.
        """
        try:
            with urlopen(  # noqa: S310
                self.repo_url,
                context=ssl.create_default_context(cafile=certifi.where()),
            ) as repo_data:
                remote_repo = json.load(repo_data)

            if config.REPO_DATA_FILE_EXISTS:
                old_hash = config.REPO_DATA["hash"]
                old_time = config.REPO_DATA["hash_utc"]

                remote_hash = remote_repo["hash"]
                remote_time = remote_repo["hash_utc"]

                if old_hash != remote_hash and old_time < remote_time:
                    config.REPO_DATA = remote_repo
            else:
                config.REPO_DATA = remote_repo
                config.REPO_DATA_FILE_EXISTS = True
                self._write_repo_data_file()
        except SSLCertVerificationError:
            print(  # noqa: T201
                """[red]Unable to fetch repo list from github.
                        - Please install certifi package with:
                            pip install certifi`"""  # noqa: W605
            )
        except URLError:
            # pylint: disable=anomalous-backslash-in-string
            print(  # noqa:T201
                """[red]Unable to fetch repo list from github.
                - You can manually update using `-U \[pair(s)]`"""  # noqa:W605
            )

    def get_available_repo_data(self) -> dict | None:
        """Fetch available data based on -p Pairs filter.

        Raises:
            SystemExit: Raises only if called from the cli.

        Returns:
            dict: If called from API, returns dict of the form:
              {'pair': {'start': 'datemonth', 'end': 'datemonth'}
              ...}
        """
        filter_pairs = config.ARGS["pairs"] - set(config.REPO_DATA)
        config.FILTER_PAIRS = None if len(filter_pairs) == 0 else filter_pairs

        if self.check_if_queue_is_needed():
            scraper = Scraper()
            scraper.populate_initial_queue()

        if self._check_for_create_or_update():
            scraper.validate_urls()
            self._write_repo_data_file()

        if config.ARGS["from_api"]:
            return self._sort_repo_dict_by(
                config.REPO_DATA.copy(), config.ARGS["pairs"]
            )

        self._print_repo_data_table()
        raise SystemExit(0)

    def _check_for_create_or_update(self) -> bool:
        """Conditions for creating or updating repo data.

        Returns:
            bool: conditonal filter
        """
        return bool(
            config.ARGS["update_remote_data"]
            or not config.REPO_DATA_FILE_EXISTS
            or config.FILTER_PAIRS
        )

    def _write_repo_data_file(self) -> None:
        """Write repository data file with hash. Create directories if needed."""
        self._hash_repo()

        create_full_path(self.repo_local_path.parent)

        with self.repo_local_path.open("w", encoding="UTF-8") as target:
            json.dump(config.REPO_DATA, target)

    def _hash_repo(self) -> None:
        """Sanitize global data repo and update hash and timestamp."""
        if "hash" in config.REPO_DATA:
            config.REPO_DATA.pop("hash", None)
        if "hash_utc" in config.REPO_DATA:
            config.REPO_DATA.pop("hash_utc", None)
        config.REPO_DATA["hash"] = hash_dict(config.REPO_DATA)
        config.REPO_DATA["hash_utc"] = get_now_utc_timestamp()

    def _sort_repo_dict_by(
        self, repo_dict_copy: dict, filter_pairs: set
    ) -> dict:  # noqa: LN001
        # pylint: disable=line-too-long
        """Sorts the output/return according to argument "--by".

            Pairs (alpha) ascending    - pair_asc
            Pairs (alpha) descending   - pair_dsc
            data start date ascending  - start_asc
            data start date descending - start_dsc

        Args:
            repo_dict_copy (dict): a copy of the global date range repository
            filter_pairs (set): set derived from CLI arg '-p' or API option
                                'pairs'

        Returns:
            dict: returns sorted dict of the form:
              {'pair': {'start': 'datemonth', 'end': 'datemonth'}
              ...}
        """
        filtered_pairs: dict = self._filter_repo_dict_by_pairs(
            repo_dict_copy, filter_pairs
        )

        match config.ARGS["by"]:
            case "pair_asc":
                filtered_pairs = dict(sorted(filtered_pairs.items()))
            case "pair_dsc":
                filtered_pairs = dict(  # noqa:BLK100
                    sorted(filtered_pairs.items(), reverse=True)
                )
            case "start_asc":
                filtered_pairs = dict(
                    sorted(
                        filtered_pairs.items(), key=lambda pair: pair[1]["start"]  # type: ignore # noqa: E501,LN002
                    )
                )
            case "start_dsc":
                filtered_pairs = dict(
                    sorted(
                        filtered_pairs.items(),
                        key=lambda pair: pair[1]["start"],  # type: ignore
                        reverse=True,
                    )
                )
        return filtered_pairs

    def _filter_repo_dict_by_pairs(
        self, repo_dict_copy: dict, filter_pairs: set
    ) -> dict:
        """Filter repo dict data according to single pair values xxxyyy.

        Args:
            repo_dict_copy (dict): a copy of the global date range repository
            filter_pairs (set): set derived from CLI arg '-p' or API option
                                'pairs'

        Returns:
            dict: returns dict created from '-p pairs' set it the form:
              {'pair': {'start': 'datemonth', 'end': 'datemonth'}
              ...}
        """
        filtered: dict = {
            pair: {
                "start": repo_dict_copy[pair]["start"],
                "end": repo_dict_copy[pair]["end"],
            }
            for pair in set(repo_dict_copy) & filter_pairs
        }

        return filtered if filter_pairs else repo_dict_copy

    def _print_repo_data_table(self) -> None:
        """Print filtered repo info to terminal."""
        table = Table(
            title="Data and date ranges available from HistData.com",
            box=box.MARKDOWN,
        )
        table.add_column("Pair -p")
        table.add_column("Start -s")
        table.add_column("End -e")

        for row in self._sort_repo_dict_by(  # pylint: disable=not-an-iterable
            config.REPO_DATA.copy(),
            config.ARGS["pairs"],
        ):
            start = config.REPO_DATA[row]["start"]
            start_year = get_year_from_datemonth(start)
            start_month = get_month_from_datemonth(start)
            end = config.REPO_DATA[row]["end"]
            end_year = get_year_from_datemonth(end)
            end_month = get_month_from_datemonth(end)
            table.add_row(
                row.lower(),
                f"{start_year}-{start_month}",
                f"{end_year}-{end_month}",
            )
        print(table)  # noqa: T201



"""Scrape histdata.com website for pairs data.

Raises:
    ValueError: On stale download url
    SystemExit: On any undefined error from scraping
"""

import os
import sys
import traceback
from pathlib import Path
from typing import Callable

import bs4
import requests
from bs4 import BeautifulSoup
from rich import print  # pylint: disable=redefined-builtin
from rich.progress import Progress, SpinnerColumn, TextColumn, TimeElapsedColumn

from histdatacom import config
from histdatacom.concurrency import ThreadPool, get_pool_cpu_count
from histdatacom.records import Record
from histdatacom.scraper.urls import Urls


class Scraper:  # noqa:H601
    """Scrape histdata.com website for pairs data.

    Attributes:
        set_repo_datum: static method from scraper.repo.Repo
        check_if_queue_is_needed: static method from scraper.repo.Repo
        check_for_repo_action: static method from scraper.repo.Repo

    Raises:
        ValueError: On stale download url
        SystemExit: On any undefined error from scraping
    """

    def __init__(self) -> None:
        """Initialize parameters for requests."""
        # pylint: disable-next=import-outside-toplevel
        from histdatacom.scraper.repo import Repo  # noqa:WPS131

        self.set_repo_datum: Callable = Repo.set_repo_datum
        self.check_if_queue_is_needed: Callable = Repo.check_if_queue_is_needed
        self.check_for_repo_action: Callable = Repo.check_for_repo_action

        # Setup
        self.urls = Urls()
        self._ensure_pairs()

    @classmethod
    def get_zip_file(cls, record: Record) -> None:
        """Download and write zip file to disk.

        Args:
            record (Record): a record from the work queue
        """
        res = cls._request_file(record, config.REQUESTS_TIMEOUT)
        record.zip_filename = cls._get_zip_file_name(res)
        cls._write_file(record, res.content)

    @classmethod
    def _get_zip_file_name(cls, response: requests.Response) -> str:
        """Parse the content-disposition header and return the zip file name.

        Args:
            response (requests.Response): Response

        Returns:
            str: *.zip file name
        """
        content_disposition_header = response.headers["Content-Disposition"]
        return str(content_disposition_header.split(";")[1].split("=")[1])

    @classmethod
    def _write_file(cls, record: Record, zip_content: bytes) -> None:
        """Write binary zip data to disk.

        Args:
            record (Record): a record from the work queue.
            zip_content (bytes): binary zip data
        """
        zip_path = Path(record.data_dir, record.zip_filename)
        zip_path.write_bytes(zip_content)

    @classmethod
    def _request_file(cls, record: Record, timeout: int) -> requests.Response:
        """Place a POST request for zip file to http://www.histdata.com/get.php.

        Args:
            record (Record): a record from the work queue
            timeout (int): retry timeout for POST

        Returns:
            requests.Response: zip file response
        """
        post_headers = config.POST_HEADERS
        post_headers["Referer"] = record.url
        return requests.post(
            "http://www.histdata.com/get.php",
            data={
                "tk": record.data_tk,
                "date": record.data_date,
                "datemonth": record.data_datemonth,
                "platform": record.data_format,
                "timeframe": record.data_timeframe,
                "fxpair": record.data_fxpair,
            },
            headers=post_headers,
            timeout=timeout,
        )

    def populate_initial_queue(self) -> None:
        """Fill Current Queue with records to be acted on."""
        with Progress(
            TextColumn(text_format="[cyan] Generating API Requests"),
            SpinnerColumn(),
            SpinnerColumn(),
            SpinnerColumn(),
            TimeElapsedColumn(),
        ) as progress:
            progress.add_task("waiting", total=0)

            for url in self.urls.generate_form_urls(
                config.ARGS["start_yearmonth"],
                config.ARGS["end_yearmonth"],
                config.ARGS["formats"],
                config.FILTER_PAIRS,
                config.ARGS["timeframes"],
            ):
                record = self._init_record(url)

                if record.status != "URL_NO_REPO_DATA":
                    record.write_memento_file(  # noqa:BLK100
                        base_dir=config.ARGS["default_download_dir"]
                    )
                    if (  # noqa:BLK100
                        self.check_if_queue_is_needed()  # noqa:BLK100
                        and record.status != "URL_NEW"
                    ):
                        self.set_repo_datum(record)
                    config.NEXT_QUEUE.put(record)  # type: ignore

            config.NEXT_QUEUE.dump_to_queue(config.CURRENT_QUEUE)  # type: ignore

    def validate_urls(self) -> None:
        """Initialize and Execute a thread pool to validate generated URLs."""
        pool = ThreadPool(
            self._validate_url,
            config.ARGS,
            "Validating",
            "URLs...",
            get_pool_cpu_count(config.ARGS["cpu_utilization"]) * 3,
        )

        pool(config.CURRENT_QUEUE, config.NEXT_QUEUE)

    def download_zips(self) -> None:
        """Initialize and Execute a thread pool to download zip archives."""
        pool = ThreadPool(
            self._download_zip,
            config.ARGS,
            "Downloading",
            "ZIPs...",
            get_pool_cpu_count(config.ARGS["cpu_utilization"]) * 3,
        )

        pool(config.CURRENT_QUEUE, config.NEXT_QUEUE)

    def _init_record(self, url: str) -> Record:
        """Create a new record for processing.

        Create a placeholder record, and if the record is already on disk,
        restore it.

        Args:
            url (str): url as primary ID of record.

        Returns:
            record (Record): a record for the work queue.
        """
        record = Record()
        record(url=url, status="URL_NEW")
        record.restore_momento(base_dir=config.ARGS["default_download_dir"])
        return record

    def _ensure_pairs(self) -> None:
        """Normalize pairs input for initial queue."""
        if (
            not (
                config.ARGS["update_remote_data"]
                and config.ARGS["available_remote_data"]
            )
            and config.FILTER_PAIRS is None
        ):
            config.FILTER_PAIRS = config.ARGS["pairs"]

    def _validate_url(self, record: Record, args: dict) -> None:
        """Scrape url for the presence of downloadable zips and related metadata.

        executed by the validate_urls thread pool.

        Args:
            record (Record): a record from the work queue.
            args (dict): a global config.ARGS dict.

        Raises:
            SystemExit: On any undefined error from scraping
        """
        try:
            if record.status == "URL_NEW":
                record = self._scrape_record_info(record)
                self._check_for_valid_download(record)

                if self.check_if_queue_is_needed():
                    self.set_repo_datum(record)

                record.status = "URL_VALID"
                record.write_memento_file(base_dir=args["default_download_dir"])

            config.NEXT_QUEUE.put(record)  # type: ignore
        except ValueError:
            if not self.check_for_repo_action():
                print(  # noqa:T201,BLK100
                    f"Info: Histdata.com does not have: {record.url}"
                )

            record.status = "URL_NO_REPO_DATA"
            record.write_memento_file(base_dir=args["default_download_dir"])
        except Exception as err:
            print(  # noqa:T201
                f"Unknown Error for URL: {record.url}",
                err,
                traceback.format_exc(),
            )
            record.delete_momento_file()
            raise SystemExit from err
        finally:
            config.CURRENT_QUEUE.task_done()  # type: ignore

    def _scrape_record_info(self, record: Record) -> Record:
        """Scrape page for archive meta data and populate record with info.

        Args:
            record (Record): a Record with a url string in Record.url

        Returns:
            Record: a record for the work queue.
        """
        page_data: dict = self._get_page_data(  # noqa:BLK001
            record.url, config.REQUESTS_TIMEOUT
        )
        self._fetch_form_values(page_data, record)
        return record

    def _check_for_valid_download(self, record: Record) -> None:
        """Check for the existence of the tk (token) key.

        Args:
            record (Record): a record with page metadata

        Raises:
            ValueError: If the token (tk) has no data.
        """
        if record.data_tk == "":
            raise ValueError

    def _download_zip(self, record: Record, args: dict) -> None:
        """Download zip from record.url.

        Executed by the download_zips thread pool.

        Args:
            record (Record): a record from the queue.
            args (dict): a global config.ARGS dict.

        Raises:
            KeyError: Invalid Zip from remote. # noqa:DAR402
            Exception: Unknown error.
        """
        try:
            if "URL_VALID" in record.status or (
                args["from_api"]
                and not self._check_for_existing_archives_on_disk(record)
            ):

                self.get_zip_file(record)
                record.status = (
                    "CSV_ZIP" if record.status == "URL_VALID" else record.status
                )
                record.write_memento_file(base_dir=args["default_download_dir"])

            config.NEXT_QUEUE.put(record)  # type: ignore
        except KeyError:
            print(  # noqa:T201
                f"Invalid Zip on histdata.com: {record.url}",
                sys.exc_info(),
            )
            record.delete_momento_file()
        except Exception:
            print("Unexpected error:", sys.exc_info())  # noqa:T201
            record.delete_momento_file()
            raise
        finally:
            config.CURRENT_QUEUE.task_done()  # type: ignore

    def _check_for_existing_archives_on_disk(self, record: Record) -> bool:
        """Check for zip, csv, or jay file.

        Args:
            record (Record): a record from the work queue.

        Returns:
            bool: file exists.
        """
        return bool(
            os.path.exists(record.data_dir + record.zip_filename)
            or os.path.exists(record.data_dir + record.csv_filename)
            or os.path.exists(record.data_dir + record.jay_filename)
        )

    def _get_page_data(self, url: str, timeout: int) -> dict:
        """Get the whole page's html.

        Args:
            url (str): url of the archive record.
            timeout (int): requests' timeout

        Returns:
            dict: of the form:
                {"page_content": html_content,
                 "encoding": encoding,
                 "bytes_length: bytes_length}
        """
        request = requests.get(url, timeout=timeout)

        page_content = BeautifulSoup(request.content, "html.parser")
        encoding = dict(request.headers)["Content-Encoding"]
        bytes_length = dict(request.headers)["Content-Length"]

        return {
            "page_content": page_content,
            "encoding": encoding,
            "bytes_length": bytes_length,
        }

    def _fetch_form_values(self, page_data: dict, record: Record) -> Record:
        """Get values from page's file download form.

        Args:
            page_data (dict): dict from _get_page_data()
            record (Record): record from the work queue.

        Returns:
            Record: record for the work queue.
        """
        form_page_content = page_data["page_content"]
        form = form_page_content.find_all("form", id="file_down")

        for element in form:
            for form_value in element:
                if isinstance(form_value, bs4.element.Tag):
                    match form_value.get("id"):
                        case "tk":
                            record.data_tk = form_value.get("value")
                        case "date":
                            record.data_date = form_value.get("value")
                        case "datemonth":
                            record.data_datemonth = form_value.get("value")
                        case "platform":
                            record.data_format = form_value.get("value")
                        case "timeframe":
                            record.data_timeframe = form_value.get("value")
                        case "fxpair":
                            record.data_fxpair = form_value.get("value")
        return record



"""Generate histdata.com archive urls.

Yields:
    url (str): histdata.com archive url
"""
from typing import Any, Generator, Optional, Set

from histdatacom.fx_enums import Timeframe, get_valid_format_timeframes
from histdatacom.utils import (
    get_current_datemonth_gmt_minus5,
    get_year_from_datemonth,
    get_month_from_datemonth,
)


class Urls:  # noqa:H601
    """Generate histdata.com archive urls.

    Yields:
        url (str): histdata.com archive url
    """

    def __init__(self) -> None:
        """Set base_url for histdata.com archives."""
        self.base_url: str = "http://www.histdata.com/download-free-forex-data/"

    def generate_form_urls(  # noqa:CCR001
        self,
        start_yearmonth: str,
        end_yearmonth: str,
        formats: set,
        pairs: Optional[Set[Any]],
        timeframes: set,
    ) -> Generator[str, None, None]:
        """Generate histdata.com urls.

        Generates permutations of user settings to yield valid urls
        for scraping.

        Args:
            start_yearmonth (str): numeric string - YYYYMM
            end_yearmonth (str): numeric string - YYYYMM
            formats (set): fx_enums.Format
            pairs (Optional[Set[Any]]): fx_enums.Pairs
            timeframes (set): fx_enums.Timeframe

        Yields:
            Generator[str, None, None]: _description_
        """
        current_yearmonth = get_current_datemonth_gmt_minus5()
        current_year = int(get_year_from_datemonth(current_yearmonth))

        if start_yearmonth is None and end_yearmonth is None:  # type: ignore
            # pylint: disable-next=line-too-long
            start_yearmonth, end_yearmonth = "200001", current_yearmonth  # type: ignore # noqa:LN002

        for sub_url, timeframe in self._valid_format_timeframe_pair_urls(
            formats, timeframes, pairs
        ):
            form_url = f"{self.base_url}?/{sub_url}"

            if end_yearmonth is None:
                for date_url in self._yield_single_year_or_month(  # type: ignore
                    timeframe, start_yearmonth
                ):
                    yield f"{form_url}{date_url}"
            else:
                start_year = int(get_year_from_datemonth(start_yearmonth))
                start_month = int(  # noqa:BLK100
                    get_month_from_datemonth(start_yearmonth)
                )
                end_year = int(get_year_from_datemonth(end_yearmonth))
                end_month = int(get_month_from_datemonth(end_yearmonth))

                # pylint: disable=not-an-iterable
                for year in range(start_year, end_year + 1):
                    yield from self._yield_range_of_yearmonths(
                        year,
                        timeframe,
                        form_url,
                        start_year,
                        start_month,
                        end_year,
                        end_month,
                        current_year,
                    )

    def _valid_format_timeframe_pair_urls(  # noqa:CCR001
        self, formats: set, timeframes: set, pairs: Optional[Set[Any]]
    ) -> Generator[tuple[str, Any], None, None]:
        """Yield permutations of format, timeframe, and pair as url fragments.

        Args:
            formats (set): fx_enums.Format
            pairs (Optional[Set[Any]]): fx_enums.Pairs
            timeframes (set): fx_enums.Timeframe

        Yields:
            Generator[tuple[str, Any], None, None]: {format}/{timeframe}/{pair}/
        """
        for csv_format in formats:
            for timeframe in timeframes:
                if timeframe in get_valid_format_timeframes(csv_format):
                    for pair in pairs:  # type: ignore
                        yield (
                            f"{csv_format}/"  # noqa:BLK001
                            f"{Timeframe[timeframe].value}/"
                            f"{pair}/"
                        ), timeframe

    def _yield_range_of_yearmonths(  # noqa:CFQ002
        self,
        year: int,
        timeframe: str,
        form_url: str,
        start_year: int,
        start_month: int,
        end_year: int,
        end_month: int,
        current_year: int,
    ) -> Generator[str, None, None]:
        """Generate final urls.

        Generate final urls, handling year/month edge cases.

        Args:
            year (int): current year in range.
            timeframe (str): fx_enums.Timeframe
            form_url (str): url fragment {base_url}/{format}/{timeframe}/{pair}
            start_year (int): YYYY
            start_month (int): MM
            end_year (int): YYYY
            end_month (int): MM
            current_year (int): YYYY

        Yields:
            Generator[str, None, None]:
                {base_url}/{format}/{timeframe}/{pair}/{year}/{month}
        """
        match year:
            case _ if year == current_year:
                for date_url in self._yield_current_year(
                    year, start_year, start_month, end_year, end_month
                ):
                    yield f"{form_url}{date_url}"

            case _ if start_year == year == end_year:
                for (
                    date_url
                ) in self._yield_same_year(  # pylint: disable=not-an-iterable
                    timeframe, year, start_month, end_month
                ):
                    yield f"{form_url}{date_url}"

            case _ if year == start_year != end_year:
                for (
                    date_url
                ) in self._yield_start_year(  # pylint: disable=not-an-iterable
                    timeframe, year, start_month
                ):
                    yield f"{form_url}{date_url}"

            case _ if year == end_year != start_year:
                for (  # noqa:BLK001
                    date_url
                ) in self._yield_end_year(  # pylint: disable=not-an-iterable
                    timeframe, year, end_month
                ):
                    yield f"{form_url}{date_url}"
            case _:
                for (  # noqa:BLK001
                    date_url
                ) in self._yield_year(  # pylint: disable=not-an-iterable
                    timeframe, year
                ):
                    yield f"{form_url}{date_url}"

    def _yield_current_year(
        self,
        year: int,
        start_year: int,
        start_month: int,
        end_year: int,
        end_month: int,
    ) -> Generator[str, None, None]:
        """Handle edge case for current/incomplete year.

        Args:
            year (int): YYYY
            start_year (int): YYYY
            start_month (int): YYYY
            end_year (int): YYYY
            end_month (int): YYYY

        Yields:
            Generator[str, None, None]: url fragment {year}/{month}
        """
        if start_year == end_year:
            for month in range(start_month, end_month + 1):
                yield f"{year}/{month}"
        else:
            for month in range(1, end_month + 1):
                yield f"{year}/{month}"

    def _yield_same_year(
        self, timeframe: str, year: int, start_month: int, end_month: int
    ) -> Generator[str, None, None]:
        """Handle edge case for same/incomplete year and M1/Tick differences.

        Args:
            timeframe (str): fx_enums.Timeframe
            year (int): YYYY
            start_month (int): MM
            end_month (int): MM

        Yields:
            Generator[str, None, None]: url fragment
                - M1: {year}
                - Tick:{year}/{month}
        """
        match timeframe:
            case "M1":
                yield f"{year}"
            case _:
                for month in range(start_month, end_month + 1):
                    yield f"{year}/{month}"

    def _yield_start_year(
        self, timeframe: str, year: int, start_month: int
    ) -> Generator[str, None, None]:
        """Handle edge case for start/incomplete year and M1/Tick differences.

        Args:
            timeframe (str): fx_enums.Timeframe
            year (int): YYYY
            start_month (int): MM

        Yields:
            Generator[str, None, None]: url fragment
                - M1: {year}
                - Tick:{year}/{month}
        """
        match timeframe:
            case "M1":
                yield f"{year}"
            case _:
                for month in range(start_month, 12 + 1):
                    yield f"{year}/{month}"

    def _yield_end_year(
        self, timeframe: str, year: int, end_month: int
    ) -> Generator[str, None, None]:
        """Handle edge case for end/incomplete year and M1/Tick differences.

        Args:
            timeframe (str): fx_enums.Timeframe
            year (int): YYYY
            end_month (int): MM

        Yields:
            Generator[str, None, None]: url fragment
                - M1: {year}
                - Tick:{year}/{month}
        """
        match timeframe:
            case "M1":
                yield f"{year}"
            case _:
                for month in range(1, end_month + 1):
                    yield f"{year}/{month}"

    def _yield_year(  # noqa:BLK100
        self, timeframe: str, year: int
    ) -> Generator[str, None, None]:
        """Handle edge case M1/Tick differences.

        On year's end, histdata.com concatenates M1 monthly data into
        a single file. This means the current year has n monthly files
        where n is the number of months upto the current, but all previous
        years are single file which is indexed by year (in the url). Tick
        data, on the other hand, is month-by-month regardless of the year.

        Args:
            timeframe (str): fx_enums.Timeframe
            year (int): YYYY

        Yields:
            Generator[str, None, None]: url fragment
                - M1: {year}
                - Tick:{year}/{month}
        """
        match timeframe:
            case "M1":
                yield f"{year}"
            case _:
                for month in range(1, 12 + 1):
                    yield f"{year}/{month}"

    def _yield_single_year_or_month(  # noqa:CCR001
        self, timeframe: str, start_yearmonth: int
    ) -> Generator[str, None, None]:
        """Handle edge case where user setting is for a single year or month.

           handles a range-less edge case.

        Args:
            timeframe (str): _description_
            start_yearmonth (int): _description_

        Yields:
            Generator[str, None, None]: url fragment
                - M1: {year}
                - Tick:{year}/{month}
        """
        current_yearmonth = get_current_datemonth_gmt_minus5()
        current_year = int(get_year_from_datemonth(current_yearmonth))
        current_month = int(get_month_from_datemonth(current_yearmonth))

        start_year = int(get_year_from_datemonth(start_yearmonth))
        start_month = int(get_month_from_datemonth(start_yearmonth))

        if start_month == 0:  # sourcery skip:merge-else-if-into-elif
            if start_year == current_year:
                for month in range(1, current_month + 1):
                    yield f"{start_year}/{month}"
            else:
                yield from self._yield_year(  # pylint: disable=not-an-iterable
                    timeframe, start_year
                )
        else:
            if start_year == current_year:
                yield f"{start_year}/{start_month}"
            else:
                match timeframe:
                    case "M1":
                        yield f"{start_year}"
                    case _:
                        yield f"{start_year}/{start_month}"



"""Utilities for histdatacom."""
from __future__ import annotations

import hashlib
import json
import os
from pathlib import Path
import re
from datetime import datetime
from typing import Any, Optional
import sys
import importlib

import pytz
import yaml
from rich import print  # pylint: disable=redefined-builtin


def get_month_from_datemonth(
    datemonth: str | int,
) -> str:
    """Extract month from datemonth string.

    Args:
        datemonth (Optional[str] | Optional[int]): YYYYMM

    Returns:
        str: MM
    """
    return (
        datemonth[-2:]  # type: ignore
        if datemonth is not None and len(datemonth) > 4  # type: ignore
        else ""
    )  # noqa: E501


def get_year_from_datemonth(
    datemonth: Optional[str] | Optional[int],
) -> str:
    """Extract year from datemonth string.

    Args:
        datemonth (Optional[str] | Optional[int]): YYYYMM

    Returns:
        str: YYYY
    """
    return datemonth[:4] if datemonth is not None else ""  # type: ignore


def force_datemonth_if_only_year(datemonth: str) -> str:
    """Rewrite datemonth if no month provided.

    Args:
        datemonth (str): YYYY

    Returns:
        str: YYYYMM
    """
    return f"{datemonth}01" if len(datemonth) == 4 else datemonth


def get_query_string(url: str) -> list[str]:
    """Extract query string from histdata.com archive urls.

    Args:
        url (str): archive url.

    Returns:
        list[str]: query string, url after '?'
    """
    return url.split("?")[1].split("/")


def create_full_path(path_str: str | Path) -> None:
    """Check if path exists. If not, create it.

    Args:
        path_str (str): path
    """
    path = Path(path_str)
    if not path.exists():
        path.mkdir(parents=True)


def set_working_data_dir(data_dir_name: str) -> str:
    """Create current working directory str.

    Args:
        data_dir_name (str): dir string

    Returns:
        str: working dir string
    """
    return f"{Path.cwd()}{os.sep}{data_dir_name}{os.sep}"


def load_influx_yaml() -> dict | Any:
    """Load settings from influx.yaml.

    # noqa: DAR402

    Raises:
        yaml.YAMLError: malformed yaml file.
        SystemExit: Exit on error

    Returns:
        yaml_file (dict): influx.yaml as dict
    """
    yaml_path = Path(".", "influxdb.yaml")

    if yaml_path.exists():
        with yaml_path.open("r", encoding="UTF-8") as file:
            try:
                yaml_file = yaml.safe_load(file)
            except yaml.YAMLError as exc:
                raise SystemExit from exc

        return yaml_file

    print(  # noqa:T201
        """ ERROR: -I flag is used to import data to a influxdb instance...
                        there is no influxdb.yaml file in working directory.
                        did you forget to set it up?
            """
    )
    raise SystemExit


def get_current_datemonth_gmt_minus5() -> str:
    """Adjust the current date month to ESTnoDST.

    Returns:
        str: YYYYMM
    """
    now: datetime = datetime.now().astimezone()
    gmt_minus5: datetime = now.astimezone(pytz.timezone("Etc/GMT-5"))
    return f"{gmt_minus5.year}{gmt_minus5.strftime('%m')}"


def replace_date_punct(datemonth_str: Optional[str]) -> str:
    """Remove year-month punctuation and returns str("000000").

    Args:
        datemonth_str (Optional[str]): YYYYMM

    Returns:
        str: YYYYMM
    """
    return (  # noqa:BLK100
        re.sub("[-_.: ]", "", datemonth_str)  # noqa:BLK100
        if datemonth_str is not None
        else ""
    )


def hash_dict(data_dict: dict) -> str:
    """Generate a MD5 identity hash for a dict.

    Args:
        data_dict (dict): dict to hash

    Returns:
        str: MD5 identity hash
    """
    dict_hash = hashlib.md5(usedforsecurity=False)
    encoded = json.dumps(data_dict, sort_keys=True).encode()
    dict_hash.update(encoded)
    return dict_hash.hexdigest()


def get_now_utc_timestamp() -> float:
    """Get the current timestamp in UTC.

    Returns:
        float: UTC timestamp
    """
    return datetime.utcnow().timestamp()  # sourcery skip


def check_installed_module(  # noqa:CCR001,BLK100
    module_name: str, load: bool = False
) -> bool:
    """Check to see if module is installed.

    Args:
        module_name (str): a module name
        load (bool): load module into sys.modules

    Raises:
        SystemExit: exit on error
        SystemExit: exit on error

    Returns:
        bool: True module is installed and available
    """
    if module_name == "arrow":
        module_name = "pyarrow"

    if module_name in sys.modules:  # noqa:R505
        return True
    elif (  # noqa:BLK100
        spec := importlib.util.find_spec(module_name)  # type: ignore
    ) is not None:
        module = importlib.util.module_from_spec(spec)  # type: ignore
        if load:
            sys.modules[module] = module
            spec.loader.exec_module(module)  # type: ignore
        return True
    else:
        if module_name == "datatable":
            err_text = (
                "datatable not installed. please visit "
                "https://github.com/dmidlo/histdata.com-tools"
                "#datatable-installation-options "
                "for installation instructions."
            )
            raise SystemExit(err_text)

        if module_name == "pyarrow":
            module_name = "arrow"

        err_text = (
            f"{module_name} format not installed. "
            "please run:\n\n  "
            f"pip install histdatacom[{module_name}]"
            "\n\n to install."
        )

        raise SystemExit(err_text)



"""Run main application. Core logic.

Raises:
    SystemExit: Exit when complete.

Returns:
    QueueManager: multi-process serial communication manager
    repo_data (set): a set of repo pairs with start and end date ranges.
    Data (Frame | DataFrame | Table):
        a datatable Frame, a pandas DataFrame, or a pyarrow Table
    List Of Data:   [
                        {
                            "timeframe": timeframe,
                            "pair": pair,
                            "records": [record, record, ...],
                            "data": Frame | DataFrame | Table,
                        },
                        ...
                        ...
                    ]

"""
from __future__ import annotations

from typing import TYPE_CHECKING

import histdatacom
from histdatacom import Options, config
from histdatacom.cli import ArgParser
from histdatacom.concurrency import QueueManager
from histdatacom.csvs import Csv
from histdatacom.scraper.repo import Repo
from histdatacom.scraper.scraper import Scraper
from histdatacom.utils import (
    load_influx_yaml,
    set_working_data_dir,
    check_installed_module,
)

if TYPE_CHECKING:
    from datatable import Frame  # noqa:I900
    from pandas import DataFrame
    from pyarrow import Table


class _HistDataCom:  # noqa:R701
    """Pull market data from histdata.com and import it into influxDB."""

    def __init__(self, options: Options) -> None:  # noqa:CCR001
        # pylint: disable=import-outside-toplevel
        """Initialize _HistDataCom Class.

        Args:
            options (Options): from histdata.options import Options

        Set User () or Default Arguments respectively utilizing the
        self.ArgParser and self.Options classes.
          - ArgParser()():
              - ()(): use an IIFE to allow argparse to get garbage collected
              - ()(): ArgParser.__call__ returns updated Options object
              - vars(...): get the __dict__ representation of the object
              - ArgParser._arg_list_to_set(...)
                  - Normalize iterable user arguments whose values are lists and
                    make them sets instead
              - .copy(): decouple for GC using a hard copy of user args
        """
        config.ARGS = ArgParser.arg_list_to_set(  # noqa:BLK100
            vars(ArgParser(options)())  # noqa:WPS110
        ).copy()
        config.ARGS["default_download_dir"] = set_working_data_dir(
            config.ARGS["data_directory"]
        )

        if config.ARGS["import_to_influxdb"]:
            influx_yaml = load_influx_yaml()
            config.ARGS["INFLUX_ORG"] = influx_yaml["influxdb"]["org"]
            config.ARGS["INFLUX_BUCKET"] = influx_yaml["influxdb"]["bucket"]
            config.ARGS["INFLUX_URL"] = influx_yaml["influxdb"]["url"]
            config.ARGS["INFLUX_TOKEN"] = influx_yaml["influxdb"]["token"]

        self.repo = Repo()
        self.scraper = Scraper()
        self.csvs = Csv()

        if config.ARGS["from_api"] and config.ARGS["api_return_type"]:
            from histdatacom.api import Api

            check_installed_module(config.ARGS["api_return_type"])
            self.api = Api()

        if config.ARGS["import_to_influxdb"]:
            config.ARGS["api_return_type"] = "datatable"
            check_installed_module(config.ARGS["api_return_type"])
            from histdatacom.influx import Influx

            self.influx = Influx()

        if (  # noqa:BLK100
            config.ARGS["available_remote_data"]  # noqa:BLK100
            or config.ARGS["update_remote_data"]
        ):
            if self.repo.test_for_repo_data_file():
                self.repo.read_repo_data_file()
            self.repo.update_repo_from_github()

    def run(  # noqa:CCR001,CFQ004,CCR001,R701
        self,
    ) -> list | dict | Frame | DataFrame | Table | None:  # noqa:CCR001
        """Execute. histdatacom's execution order.

        Returns:
            list | dict | Frame | DataFrame | Table | None:

            Data (Frame | DataFrame | Table):
                    a datatable Frame, a pandas DataFrame, or a pyarrow Table
            List of dicts:  [
                                {
                                    "timeframe": timeframe,
                                    "pair": pair,
                                    "records": [record, record, ...],
                                    "data": Frame | DataFrame | Table,
                                },
                                ...
                                ...
                            ]


        """
        if config.ARGS["version"]:
            if not config.ARGS["from_api"]:
                print(histdatacom.__version__)  # noqa:T201
            return histdatacom.__version__

        if (  # noqa:BLK100
            config.ARGS["available_remote_data"]  # noqa:BLK100
            or config.ARGS["update_remote_data"]
        ):
            return self.repo.get_available_repo_data()
        del self.repo  # noqa:WPS100

        self.scraper.populate_initial_queue()

        if config.ARGS["validate_urls"]:
            self.scraper.validate_urls()

        if config.ARGS["download_data_archives"]:
            self.scraper.download_zips()
            del self.scraper  # noqa:WPS100
            if config.ARGS["from_api"] and config.ARGS["api_return_type"]:
                self.api.validate_jays()
                return self.api.merge_jays()

        if config.ARGS["extract_csvs"]:
            self.csvs.extract_csvs()
        del self.csvs  # noqa:WPS100

        if config.ARGS["import_to_influxdb"]:
            self.influx.import_data()

        return None


def main(
    options: Options | None = None,
) -> list | dict | Frame | DataFrame | Table:
    """Execute. Entry-point for histdatacom.

    Args:
        options (Options): a histdatacom.options Options object.

    Returns:
        list | dict | Frame | DataFrame | Table:

            Data (Frame | DataFrame | Table):
                    a datatable Frame, a pandas DataFrame, or a pyarrow Table
            List of dicts:  [
                                {
                                    "timeframe": timeframe,
                                    "pair": pair,
                                    "records": [record, record, ...],
                                    "data": Frame | DataFrame | Table,
                                },
                                ...
                                ...
                            ]
    """
    if not options:
        options = Options()
        QueueManager(options)(_HistDataCom)
        return None
    options.from_api = True
    return QueueManager(options)(_HistDataCom)


if __name__ == "__main__":
    raise SystemExit(main())



"""Extract CSVs from zip archive."""
from __future__ import annotations

import sys
import zipfile
from pathlib import Path
from typing import TYPE_CHECKING

from rich import print  # pylint: disable=redefined-builtin

from histdatacom import config
from histdatacom.concurrency import ProcessPool, get_pool_cpu_count

if TYPE_CHECKING:
    from histdatacom.records import Record, Records


class Csv:  # noqa:H601
    """Extract CSV documents from zip archives."""

    def extract_csvs(self) -> None:
        """Execute process pool with extract_csv."""
        pool = ProcessPool(
            self._extract_csv,
            config.ARGS,
            "Extracting",
            "CSVs...",
            get_pool_cpu_count(config.ARGS["cpu_utilization"]),
        )

        pool(config.CURRENT_QUEUE, config.NEXT_QUEUE)

    def _extract_csv(
        self,
        record: Record,
        args: dict,
        records_current: Records,
        records_next: Records,
    ) -> None:
        """Extract single csv file. Called by extract_csvs.

        # noqa: DAR402

        Args:
            record (Record): a record from the work queue.
            args (dict): from config.ARGS
            records_current (Records): config.CURRENT_QUEUE
            records_next (Records): config.NEXT_QUEUE

        Raises:
            OSError: OS Error.
            SystemExit: exit on error
        """
        try:
            if "CSV_ZIP" in record.status:
                zip_path = Path(record.data_dir, record.zip_filename)

                with zipfile.ZipFile(zip_path, "r") as zip_ref:
                    [record.csv_filename] = [
                        name
                        for name in zip_ref.namelist()
                        if (".csv" or ".xlsx") in name
                    ]
                    zip_ref.extract(record.csv_filename, path=record.data_dir)

                zip_path.unlink()
                record.status = "CSV_FILE"
                record.write_memento_file(base_dir=args["default_download_dir"])
            records_next.put(record)
        except OSError as err:
            print("Unexpected error:", sys.exc_info())  # noqa:T201
            record.delete_momento_file()
            raise SystemExit from err
        finally:
            records_current.task_done()



"""Setup the setuptools version of histdatacom.

Raises:
    RuntimeError: When unable to find __version__ string

Returns:
    None
"""
import codecs
from pathlib import Path, PurePath

from setuptools import find_packages, setup


def read_init_file(rel_path: str) -> str:
    """Read the contents of a file.

    Args:
        rel_path (str): a relative file path

    Returns:
        str: contents of file
    """
    init_file_path = PurePath(__file__).parent / rel_path
    with codecs.open(str(init_file_path), "r") as file_path:
        return file_path.read()


def get_version(rel_path: str) -> str:
    """Get the version number from file by looking at  "__version__ = 'x.x.x'".

    Args:
        rel_path (str): a relative file path

    Raises:
        ErrCantFindVersionString: Unable to find version string

    Returns:
        str: string in file after line __version__ delimited by " or '
    """
    for line in read_init_file(rel_path).splitlines():
        if line.startswith("__version__"):
            deliminator = '"' if '"' in line else "'"
            return line.split(deliminator)[1]
    raise ErrCantFindVersionString(rel_path)


readme = Path("README.md")
with readme.open("r", encoding="utf-8") as readme_content:
    long_description = readme_content.read()

setup(
    # basic package data
    name="histdatacom",
    python_requires=">=3.10.0",
    version=get_version(str(PurePath("src", "histdatacom", "__init__.py"))),
    description=(
        "A Multi-threaded/Multi-Process command-line utility and "
        "python package that downloads currency exchange rates from "
        "Histdata.com. Imports to InfluxDB. Can be used in Jupyter "
        "Notebooks."
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dmidlo/histdata.com-tools",
    project_urls={
        "Bug Tracker": "https://github.com/dmidlo/histdata.com-tools/issues",
    },
    author="David Midlo",
    author_email="dmidlo@gmail.com",
    license="MIT License",
    # package structure
    packages=find_packages("src"),
    package_dir={"": "src"},
    # install the RSReader executable
    entry_points={
        "console_scripts": ["histdatacom = histdatacom.histdata_com:main"],
    },  # pylint: disable=line-too-long
    install_requires=[
        "influxdb_client",
        "rich",
        "requests",
        "bs4",
        "pyyaml",
        "rx",
        "argparse",
        "pytz",
        "certifi",
    ],
    extras_require={
        "arrow": [
            "pyarrow",
        ],
        "pandas": [
            "pandas",
        ],
        "jupyter": [
            "jupyter",
            "ipywidgets",
        ],
        "dev": [
            "pytest",
            "mypy",
            "keyring",
            "sh",
            "flake8",
            "black",
            "flake8-black",
            "coverage",
            "pytest-cov",
            "pylint",
            "bandit",
            "flake8-bandit",
            "flake8-comprehensions",
            "flake8-class-attributes-order",
            "flake8-bugbear",
            "pyflakes",
            "pre-commit",
            "types-setuptools",
            "pandas-stubs",
            "types-beautifulsoup4",
            "wemake-python-styleguide",
            "flake8-simplify",
            "flake8-pie",
            "flake8-use-pathlib",
            "flake8-use-fstring",
            "flake8-print",
            "flake8-no-implicit-concat",
            "pycodestyle",
            "flake8-pytest-style",
            "Flake8-AAA",
            "flake8-spellcheck",
            "flake8-docstring-checker",
            "flake8-docstrings",
            "flake8-coding",
            "flake8-length",
            "flake8-functions",
            "flake8-expression-complexity",
            "flake8-cognitive-complexity",
            "flake8-annotations-complexity",
            "cohesion",
            "Darglint",
            "tryceratops",
            "radon",
            "pyroma",
            "vulture",
            "flake8-type-checking",
            "isort",
            "sourcery-cli",
        ],
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Environment :: MacOS X",
        "Operating System :: MacOS",
        "Environment :: Win32 (MS Windows)",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Framework :: Jupyter",
        "Topic :: Terminals",
    ],
    keywords=[
        "finance",
        "data",
        "datascience",
        "HistData.com",
        "scraper",
        "influxdb",
        "currency exchange",
        "forex",
        "fx",
        "etl",
    ],
)


class ErrCantFindVersionString(RuntimeError):  # noqa: H601
    """RuntimeError when version string is not found.

    Args:
        path (str): tested path of __version__ string
    """

    def __init__(self, path: str) -> None:
        """Call when the version is not found.

        Args:
            path (str): tested path of __version__ string
        """
        super().__init__(f"Unable to find __version__ string in:   \n {path}")



"""Download (if needed), format, and import data to influxdb."""
# pylint: disable=redefined-outer-name
from __future__ import annotations

import sys
from collections import namedtuple
from concurrent.futures import ProcessPoolExecutor
from functools import partial
from multiprocessing import Process, Queue
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterable, Optional, Tuple

import rx
from influxdb_client import InfluxDBClient, WriteOptions, WritePrecision
from influxdb_client.client.write_api import WriteType
from reactivex.scheduler import ThreadPoolScheduler  # noqa:I900
from rich import print  # pylint: disable=redefined-builtin
from rich.progress import Progress, SpinnerColumn, TextColumn, TimeElapsedColumn
from rx import operators as ops

from histdatacom import config
from histdatacom.api import Api
from histdatacom.concurrency import ProcessPool, get_pool_cpu_count

if TYPE_CHECKING:
    from histdatacom.records import Record, Records


class Influx:  # noqa:H601
    """Download (if needed), format, and import data to influxdb."""

    def import_data(self) -> None:
        """Initialize a pool of influxdb writers and...

        processes pool with a progress bar.
        """
        for _ in range(  # noqa:BLK100
            1, get_pool_cpu_count(config.ARGS["cpu_utilization"]) + 1
        ):
            writer = InfluxDBWriter(config.ARGS, config.INFLUX_CHUNKS_QUEUE)
            writer.start()

        pool = ProcessPool(
            self._import_file,
            config.ARGS,
            "Adding",
            "CSVs to influx queue...",
            get_pool_cpu_count(config.ARGS["cpu_utilization"]),
            join=False,
            dump=False,
        )

        pool(
            config.CURRENT_QUEUE,
            config.NEXT_QUEUE,
            config.INFLUX_CHUNKS_QUEUE,
        )

        with Progress(
            TextColumn(text_format="[cyan]...finishing upload to influxdb"),
            SpinnerColumn(),
            SpinnerColumn(),
            SpinnerColumn(),
            TimeElapsedColumn(),
        ) as progress:
            task_id = progress.add_task("waiting", total=0)

            config.CURRENT_QUEUE.join()  # type: ignore
            config.INFLUX_CHUNKS_QUEUE.put(None)  # type: ignore
            config.INFLUX_CHUNKS_QUEUE.join()  # type: ignore
            progress.advance(task_id, 0.75)

        print("[cyan] done.")  # noqa:T201
        config.NEXT_QUEUE.dump_to_queue(config.CURRENT_QUEUE)  # type: ignore

    def _init_counters(
        self,
        influx_chunks_queue_: Queue,
        args_: dict,
    ) -> None:
        """Initialize pool with access to these global variables.

        Args:
            influx_chunks_queue_ (Queue): ReactiveX queue
                * serialized through SyncManager
            args_ (dict): config.ARGS
        """
        # pylint: disable=global-variable-undefined
        global INFLUX_CHUNKS_QUEUE  # noqa:WPS100
        INFLUX_CHUNKS_QUEUE = influx_chunks_queue_  # type: ignore
        global ARGS  # noqa:WPS100
        ARGS = args_  # type: ignore

    def _import_file(
        self,
        record: Record,
        args: dict,
        records_current: Records,
        records_next: Records,
        influx_chunks_queue: Queue,
    ) -> None:
        """Import ASCII data to influxdb, both for csv and jay.

        Args:
            record (Record): a record from the work queue
            args (dict): config.ARGS
            records_current (Records): config.CURRENT_QUEUE
            records_next (Records): config.NEXT_QUEUE
            influx_chunks_queue (Queue): config.INFLUX_CHUNKS_QUEUE

        Raises:
            Exception: on unknown exception.
        """
        try:
            if (
                record.status != "INFLUX_UPLOAD"
                and str.lower(record.data_format) == "ascii"
            ):
                jay_path = Path(record.data_dir, ".data")
                if jay_path.exists():
                    self._import_jay(
                        record,
                        args,
                        records_current,
                        records_next,
                        influx_chunks_queue,
                    )
                elif "CSV" in record.status:
                    Api.test_for_jay_or_create(record, args)
                    self._import_jay(
                        record,
                        args,
                        records_current,
                        records_next,
                        influx_chunks_queue,
                    )

            record.status = "INFLUX_UPLOAD"
            record.write_memento_file(base_dir=args["default_download_dir"])

            if args["delete_after_influx"]:
                Path(record.data_dir, record.zip_filename).unlink()
                Path(record.data_dir, record.jay_filename).unlink()
            records_next.put(record)
        except Exception as err:
            print(  # noqa:T201
                "Unexpected error from here:", sys.exc_info(), err
            )  # noqa:T201
            record.delete_momento_file()
            raise
        finally:
            records_current.task_done()

    def _import_jay(
        self,
        record: Record,
        args: dict,
        records_current: Records,  # noqa:W0613 # pylint: disable=W0613
        records_next: Records,  # noqa:W0613 # pylint: disable=W0613
        influx_chunks_queue: Queue,
    ) -> None:
        """Import a jay file with a ReactiveX pub/sub queue.

        Args:
            record (Record): a record from the work queue config.CURRENT_QUEUE
            args (dict): config.ARGS
            records_current (Records): config.CURRENT_QUEUE
            records_next (Records): config.NEXT_QUEUE
            influx_chunks_queue (Queue): config.INFLUX_CHUNKS_QUEUE
        """
        jay = Api.import_jay_data(record.data_dir + record.jay_filename)

        with ProcessPoolExecutor(
            max_workers=1,
            initializer=self._init_counters,
            initargs=(influx_chunks_queue, config.ARGS),
        ) as executor:

            rx_data_queue = rx.from_iterable(jay.to_tuples()).pipe(
                ops.buffer_with_count(args["batch_size"]),
                ops.flat_map(
                    lambda rows: executor.submit(  # noqa:BLK100
                        self._parse_jay_rows, rows, record
                    )
                ),
            )

            rx_data_queue.subscribe(
                on_next=lambda x: None,
                on_error=lambda er: print(  # noqa:T201
                    f"Unexpected error: {er}"
                ),  # noqa:T201
            )

    def _parse_jay_rows(self, iterable: Iterable, record: Record) -> None:
        """Create a list by mapping row-by-row from datatable Frame (from jay).

        Args:
            iterable (Iterable): datatable.Frame
            record (Record): a record from the work queue.
        """
        map_func = partial(self._parse_jay_row, record=record)
        parsed_rows = list(map(map_func, iterable))

        INFLUX_CHUNKS_QUEUE.put(parsed_rows)  # type: ignore

    def _parse_jay_row(self, row: Tuple[Any], record: Record) -> str:
        """Return influxdb line-protocol line (str) for each from a map function.

            Applies different fields for line in line-protocol on Timeframe
                M1 or T.

        Args:
            row (Tuple[Any]): row from datatable.Frame
            record (Record): record from the work queue

        Returns:
            str: line-protocol (influxdb)
        """
        # pylint: disable=line-too-long
        measurement = f"{record.data_fxpair}"
        tags = (
            f"source=histdata.com,format={record.data_format}"
            f",timeframe={record.data_timeframe}"
        ).replace(" ", "")

        match record.data_timeframe:
            case "M1":
                _row = namedtuple(
                    "_row",
                    ["datetime", "open", "high", "low", "close", "vol"],
                )
                named_row = _row(
                    row[0],
                    row[1],  # type: ignore
                    row[2],  # type: ignore
                    row[3],  # type: ignore
                    row[4],  # type: ignore
                    row[5],  # type: ignore
                )

                fields = (
                    f"openbid={named_row.open},"
                    f"highbid={named_row.high},"
                    f"lowbid={named_row.low},"
                    f"closebid={named_row.close}"
                ).replace(" ", "")
                time = str(named_row.datetime)
            case "T":
                _row = namedtuple(  # type: ignore
                    "_row", ["datetime", "bid", "ask", "vol"]
                )
                named_row = _row(row[0], row[1], row[2], row[3])  # type: ignore

                fields = (
                    f"bidquote={named_row.bid},"  # type: ignore
                    f"askquote={named_row.ask}"
                ).replace(" ", "")
                time = str(named_row.datetime)

        # return in line-protocol format.
        return f"{measurement},{tags} {fields} {time}"


class InfluxDBWriter(Process):
    """Write data from the chunks queue as line-protocol to influxdb write api.

    Args:
        Process (Process): A Python Process.
    """

    def __init__(self, args: dict, influx_chunks_queue: Optional[Queue]):
        """Initialize a process for the influxdbClient write api.

        Args:
            args (dict): config.ARGS
            influx_chunks_queue (Optional[Queue]): config.INFLUX_CHUNKS_QUEUE
        """
        Process.__init__(self)
        self.args = args
        self.influx_chunks_queue = influx_chunks_queue
        self.client = InfluxDBClient(
            url=self.args["INFLUX_URL"],
            token=self.args["INFLUX_TOKEN"],
            org=self.args["INFLUX_ORG"],
            debug=False,
        )

        self.write_api = self.client.write_api(
            write_options=WriteOptions(
                write_type=WriteType.batching,
                batch_size=args["batch_size"],
                flush_interval=args["batch_size"],
            ),
            write_scheduler=ThreadPoolScheduler(
                max_workers=get_pool_cpu_count(config.ARGS["cpu_utilization"])
            ),
        )

    def run(self) -> None:
        """Process chunks from config.INFLUX_CHUNKS_QUEUE."""
        while True:
            try:
                chunk = self.influx_chunks_queue.get()  # type: ignore
            except EOFError:
                break

            if chunk is None:
                self._terminate()
                self.influx_chunks_queue.task_done()  # type: ignore
                break

            self.write_api.write(
                org=self.args["INFLUX_ORG"],
                bucket=self.args["INFLUX_BUCKET"],
                record=chunk,
                write_precision=WritePrecision.MS,
            )
            self.influx_chunks_queue.task_done()  # type: ignore

    def _terminate(self) -> None:
        """Terminate the influxdb subprocess."""
        del self.write_api  # noqa:WPS100
        del self.client  # noqa:WPS100
